#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for a dos calculation and
some helper methods to do so with AiiDA
"""
import numpy as np
from masci_tools.io.kkr_params import kkrparams
from masci_tools.io.common_functions import get_ef_from_potfile, get_Ry2eV
from masci_tools.io.common_functions import get_alat_from_bravais
from aiida import orm
from aiida.engine import WorkChain, while_, if_, ToContext, calcfunction
from aiida.common.exceptions import NotExistent
from aiida_kkr.calculations.voro import VoronoiCalculation
from aiida_kkr.workflows.dos import kkr_dos_wc
from aiida_kkr.tools import find_cluster_radius
from aiida_kkr.tools.common_workfunctions import (
    test_and_get_codenode, update_params, update_params_wf, get_inputs_voronoi
)
from aiida_kkr.tools.save_output_nodes import create_out_dict_node

__copyright__ = (
    u'Copyright (c), 2017-2018, Forschungszentrum Jülich GmbH, '
    'IAS-1/PGI-1, Germany. All rights reserved.'
)
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.13.3'
__contributors__ = u'Philipp Rüßmann'

eV2Ry = 1.0 / get_Ry2eV()


class kkr_startpot_wc(WorkChain):
    """
    Workchain  create starting potential for a KKR calculation by running
    voronoi and getting the starting DOS for first checks on the validity of the input setting.
    Starts from a structure together with a KKR parameter node.

    :param wf_parameters: (Dict), Workchain specifications
    :param options: (Dict), specifications for the computer
    :param structure: (StructureData), aiida structure node to begin
        calculation from (needs to contain vacancies, if KKR needs empty spheres)
    :param kkr: (Code)
    :param voronoi: (Code)
    :param calc_parameters: (Dict), KKR parameter set, passed on to voronoi run.

    :return result_kkr_startpot_wc: (Dict), Information of workflow results
        like Success, last result node, dos array data
    """

    _workflowversion = __version__
    _wf_default = {
        'num_rerun': 4,  # number of times voronoi+starting dos+checks is rerun to ensure non-negative DOS etc
        'fac_cls_increase':
        1.15,  # alat         # factor by which the screening cluster is increased each iteration (up to num_rerun times)
        'natom_in_cls_min': 79,  # minimum number of atoms in screening cluster
        'delta_e_min': 1.,  # eV                  # minimal distance in DOS contour to emin and emax in eV
        'threshold_dos_zero': 10**-2,  #states/eV #
        'check_dos': False,  # logical to determine if DOS is computed and checked
        'ef_set': None  # set Fermi level of starting potential to this value
    }
    _options_default = {
        'queue_name': '',  # Queue name to submit jobs to
        'resources': {
            'num_machines': 1
        },  # resources to allowcate for the job
        'max_wallclock_seconds': 60 * 60,  # walltime after which the job gets killed (gets parsed to KKR)
        'withmpi': True,  # execute KKR with mpi or without
        'custom_scheduler_commands': '',  # some additional scheduler commands
    }
    # add defaults of dos_params since they are passed onto that workflow
    _wf_default['dos_params'] = kkr_dos_wc.get_wf_defaults(silent=True)

    _wf_label = ''
    _wf_description = ''

    _kkr_default_params = kkrparams.get_KKRcalc_parameter_defaults()

    # intended to guide user interactively in setting up a valid wf_params node
    @classmethod
    def get_wf_defaults(cls, silent=False):
        """Print and return _wf_defaults dictionary.

        Can be used to easily create set of wf_parameters.
        returns _wf_defaults
        """
        if not silent:
            print(f'Version of workflow: {cls._workflowversion}')
        return cls._wf_default

    @classmethod
    def define(cls, spec):
        """Defines the outline of the workflow."""
        # Take input of the workflow or use defaults defined above
        super(kkr_startpot_wc, cls).define(spec)
        spec.input(
            'wf_parameters',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._wf_default),
            help='Parameters that control the behavior of the workflow'
        )
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._options_default),
            help='Computer options passed onto the calculations'
        )
        spec.input(
            'structure',
            valid_type=orm.StructureData,
            required=False,
            help="""
            Structure for which the starting potential should be constructed,
            not needed if parent_KKR is given (typically used to increase the
            lmax but use the output potential of the parent_KKR as starting
            potential).
            """
        )
        spec.input(
            'kkr',
            valid_type=orm.Code,
            required=False,
            help='Kkr code, only needed only if DOS is calculated.',
        )
        spec.input(
            'voronoi',
            valid_type=orm.Code,
            required=True,
            help='Voronoi code',
        )
        spec.input(
            'calc_parameters',
            valid_type=orm.Dict,
            required=False,
            help="""
            KKR-specific parameters passed onto the VoronoiCalculation (lmax etc.).
            """,
        )
        spec.input(
            'startpot_overwrite',
            valid_type=orm.SinglefileData,
            required=False,
            help="""
            Potential which can be used instead of the output potential
            Voronoi constructs.
            """,
        )
        spec.input(
            'parent_KKR',
            valid_type=orm.RemoteData,
            required=False,
            help="""
            RemoteData node of a KKR calculation which is used to overwrite
            the output potential Voronoi constructs
            (typically used to increase lmax). Cannot be used with a different
            structure in the input since the structure is extracted from the
            parent_KKR.
            """,
        )
        # define output nodes
        spec.output(
            'results_vorostart_wc',
            valid_type=orm.Dict,
            required=True,
            help='',
        )
        spec.output(
            'last_doscal_results',
            valid_type=orm.Dict,
            required=False,
            help='',
        )
        spec.output(
            'last_voronoi_results',
            valid_type=orm.Dict,
            required=False,
            help='',
        )
        spec.output(
            'last_voronoi_remote',
            valid_type=orm.RemoteData,
            required=False,
            help='',
        )
        spec.output(
            'last_params_voronoi',
            valid_type=orm.Dict,
            required=False,
            help='',
        )
        spec.output(
            'last_doscal_dosdata',
            valid_type=orm.XyData,
            required=False,
            help='',
        )
        spec.output(
            'last_doscal_dosdata_interpol',
            valid_type=orm.XyData,
            required=False,
            help='',
        )
        # definition of exit codes if the workflow needs to be terminated
        spec.exit_code(
            201,
            'ERROR_INVALID_KKRCODE',
            message='The code you provided for kkr does not use the plugin kkr.kkr',
        )
        spec.exit_code(
            202,
            'ERROR_INVALID_VORONOICODE',
            message='The code you provided for voronoi does not use the plugin kkr.voro',
        )
        spec.exit_code(
            203,
            'ERROR_VORONOI_FAILED',
            message='Voronoi calculation unsuccessful. Check inputs',
        )
        spec.exit_code(
            204,
            'ERROR_VORONOI_PARSING_FAILED',
            message='Voronoi calculation unsuccessful. Check inputs.',
        )
        spec.exit_code(
            205,
            'ERROR_VORONOI_INVALID_RADII',
            message='Voronoi calculation unsuccessful. Structure inconsistent. Maybe you need empty spheres?',
        )
        spec.exit_code(
            206,
            'ERROR_DOSRUN_FAILED',
            message='DOS run unsuccessful. Check inputs.',
        )
        spec.exit_code(
            207,
            'ERROR_BOTH_STRUCTURE_AND_PARENT_KKR_GIVEN',
            message='Can only take either structure or parent_KKR as input.',
        )
        spec.exit_code(
            208,
            'ERROR_NEITHER_STRUCTURE_NOR_PARENT_KKR_GIVEN',
            message='Need either structure or parent_KKR as input.',
        )

        # Here the structure of the workflow is defined
        spec.outline(
            # initialize workflow and check inputs
            cls.start,
            # check if another iteration is done (in case of either voro_ok, doscheck_ok is False)
            while_(cls.do_iteration_check)(
                # run voronoi calculation
                # TODO: encapsulate this in restarting mechanism (should be a base class of workflows that start calculations)
                # i.e. use base_restart_calc workchain as parent
                cls.run_voronoi,
                # check voronoi output (also sets ctx.voro_ok)
                if_(cls.check_voronoi)(
                    # create starting DOS using dos sub-workflow
                    cls.get_dos,
                    # perform some checks and set ctx.doscheck_ok accordingly
                    cls.check_dos
                )
            ),
            # collect results and return
            cls.return_results,
            cls.error_handler
        )

    def start(self):
        """
        init context and some parameters
        """
        self.report(f'INFO: started VoroStart workflow version {self._workflowversion}')

        # check input port consistency and set structure
        if 'structure' in self.inputs:
            has_struc = True
            # save structure in context
            self.ctx.structure = self.inputs.structure
        else:
            has_struc = False
        if 'parent_KKR' in self.inputs:
            has_parent_KKR = True
            # find structure from parent_KKR
            self.ctx.structure, _ = VoronoiCalculation.find_parent_structure(self.inputs.parent_KKR)
        else:
            has_parent_KKR = False

        if has_struc and has_parent_KKR:
            self.report('ERROR: caught exit code ERROR_BOTH_STRUCTURE_AND_PARENT_KKR_GIVEN')
            return self.exit_codes.ERROR_BOTH_STRUCTURE_AND_PARENT_KKR_GIVEN  # pylint: disable=no-member
        if (not has_struc) and (not has_parent_KKR):
            self.report('ERROR: caught exit code ERROR_NEITHER_STRUCTURE_NOR_PARENT_KKR_GIVEN')
            return self.exit_codes.ERROR_NEITHER_STRUCTURE_NOR_PARENT_KKR_GIVEN  # pylint: disable=no-member

        ####### init    #######

        # internal para /control para
        self.ctx.exit_code = None

        # input para
        wf_dict = self.inputs.wf_parameters.get_dict()
        options_dict = self.inputs.options.get_dict()

        # TODO: check for completeness
        if wf_dict == {}:
            wf_dict = self._wf_default
            self.report('INFO: using default wf parameter')
        if options_dict == {}:
            options_dict = self._options_default
            self.report('INFO: using default options')

        # set values, or defaults for computer options
        self.ctx.withmpi = options_dict.get(
            'withmpi',
            self._options_default['withmpi'],
        )
        self.ctx.resources = options_dict.get(
            'resources',
            self._options_default['resources'],
        )
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds',
            self._options_default['max_wallclock_seconds'],
        )
        self.ctx.queue = options_dict.get(
            'queue_name',
            self._options_default['queue_name'],
        )
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands',
            self._options_default['custom_scheduler_commands'],
        )

        # set DOS parameters
        self.ctx.dos_params_dict = wf_dict.get(
            'dos_params',
            self._wf_default['dos_params'],
        )

        # set label and description of the workflow
        self.ctx.description_wf = self.inputs.get(
            '_description',
            self._wf_description,
        )
        self.ctx.label_wf = self.inputs.get(
            '_label',
            self._wf_label,
        )

        # iterative rerunning parameters
        self.ctx.iter = 0
        self.ctx.Nrerun = wf_dict.get(
            'num_rerun',
            self._wf_default['num_rerun'],
        )

        # initialize checking booleans
        self.ctx.is_starting_iter = True
        self.ctx.doscheck_ok = True
        self.ctx.voro_ok = False
        self.ctx.check_dos = wf_dict.get(
            'check_dos',
            self._wf_default['check_dos'],
        )
        self.ctx.dos_check_fail_reason = None

        # some physical parameters that are reused
        self.ctx.nclsmin = wf_dict.get(
            'natom_in_cls_min',
            self._wf_default['natom_in_cls_min'],
        )
        self.ctx.fac_clsincrease = wf_dict.get(
            'fac_cls_increase',
            self._wf_default['fac_cls_increase'],
        )
        self.ctx.efermi = None

        # find starting cluster radius
        # find cluster radius (in alat units)
        self.ctx.r_cls = self.find_cluster_radius_alat()

        # difference in eV to emin (e_fermi) if emin (emax) are
        # larger (smaller) than emin (e_fermi)
        self.ctx.delta_e = wf_dict.get(
            'delta_e_min',
            self._wf_default['delta_e_min'],
        )
        # threshold for dos comparison (comparison of dos at emin)
        self.ctx.threshold_dos_zero = wf_dict.get(
            'threshold_dos_zero',
            self._wf_default['threshold_dos_zero'],
        )

        # set Fermi level with input value
        self.ctx.ef_set = wf_dict.get(
            'ef_set',
            self._wf_default['ef_set'],
        )

        # TODO add missing info
        # print the inputs
        self.report(
            f'INFO: use the following parameter:\n'
            f'withmpi: {self.ctx.withmpi}\n'
            f'Resources: {self.ctx.resources}\n'
            f'Walltime (s): {self.ctx.max_wallclock_seconds}\n'
            f'queue name: {self.ctx.queue}\n'
            f'scheduler command: {self.ctx.custom_scheduler_commands}\n'
            f'description: {self.ctx.description_wf}\n'
            f'label: {self.ctx.label_wf}\n'
            f'dos_params: {self.ctx.dos_params_dict}\n'
            f'Max. number of voronoi reruns: {self.ctx.Nrerun}\n'
            f'factor cluster increase: {self.ctx.fac_clsincrease}\n'
            f'default cluster radius (in alat): {self.ctx.r_cls}\n'
            f'min. number of atoms in screening cls: {self.ctx.nclsmin}\n'
            f'min. dist in DOS contour to emin/emax: {self.ctx.delta_e} eV\n'
            f'threshold where DOS is zero: {self.ctx.threshold_dos_zero} states/eV\n'
        )

        # return para/vars
        self.ctx.successful = True
        self.ctx.errors = []
        self.ctx.formula = ''

        # get kkr and voronoi codes from input
        if self.ctx.check_dos:
            try:
                test_and_get_codenode(self.inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                return self.exit_codes.ERROR_INVALID_KKRCODE  # pylint: disable=no-member
        try:
            test_and_get_codenode(self.inputs.voronoi, 'kkr.voro', use_exceptions=True)
        except ValueError:
            return self.exit_codes.ERROR_INVALID_VORONOICODE  # pylint: disable=no-member

    def run_voronoi(self):
        """Run voronoi calculation with parameters from input."""
        # incerement iteration counter
        self.ctx.iter += 1

        # increase some parameters
        # check if cluster size is actually the reason for failure
        if self.ctx.iter > 1 and self.ctx.dos_check_fail_reason not in [
            'EMIN too high',
        ]:
            self.ctx.r_cls = self.ctx.r_cls * self.ctx.fac_clsincrease

        structure = self.ctx.structure
        self.ctx.formula = structure.get_formula()
        label = f'voronoi calculation step {self.ctx.iter}'
        description = f'{self.ctx.description_wf} voronoi on {self.ctx.formula}'

        voronoicode = self.inputs.voronoi

        # get valid KKR parameters
        if self.ctx.iter > 1:
            # take value from last run to continue
            params = self.ctx.last_params
            first_iter = False
        else:
            # used input or defaults in first iteration
            first_iter = True
            if 'calc_parameters' in self.inputs:
                params = self.inputs.calc_parameters
            else:
                kkrparams_default = kkrparams()
                para_version = self._kkr_default_params[1]
                for key, val in self._kkr_default_params[0].items():
                    kkrparams_default.set_value(key, val, silent=True)
                # create Dict node
                params = orm.Dict(dict=kkrparams_default.get_dict())
                params.label = 'Defaults for KKR parameter node'
                params.description = f'defaults as defined in kkrparams of version {para_version}'
            #  set last_params accordingly (used below for provenance tracking)
            self.ctx.last_params = params

        self.report(f'INFO: input params: {params}')

        # check if RCLUSTZ is set and use setting from wf_parameters instead
        # (calls update_params_wf to keep track of provenance)
        updated_params = False
        update_list = []
        kkr_para = kkrparams()
        for key, val in params.get_dict().items():
            kkr_para.set_value(key, val, silent=True)
        set_vals = kkr_para.get_set_values()
        set_vals = [keyvalpair[0] for keyvalpair in set_vals]
        default_values = kkrparams.get_KKRcalc_parameter_defaults()[0]
        if 'RCLUSTZ' in set_vals:
            rcls_input = params.get_dict()['RCLUSTZ']
            # set r_cls by default or from input in first iteration
            if self.ctx.r_cls < rcls_input and first_iter:
                self.ctx.r_cls = rcls_input
                updated_params = True
                update_list.append('RCLUSTZ')
            elif self.ctx.nclsmin > 0 and self.ctx.r_cls > rcls_input:
                # change rcls with iterations, do this only if nclsmin is >0
                # (0 or negative numbers trigger use of RCLUSTZ)
                updated_params = True
                update_list.append('RCLUSTZ')
        else:
            updated_params = True
            update_list.append('RCLUSTZ')
        # in case of dos check verify that RMAX, GMAX are set and use setting
        # from wf_parameters otherwise
        if 'RMAX' in set_vals:
            update_list.append('RMAX')
            rmax_input = params.get_dict()['RMAX']
        elif self.ctx.check_dos:  # add only if doscheck is done
            updated_params = True
            update_list.append('RMAX')
            rmax_input = default_values.get('RMAX')
        if 'GMAX' in set_vals:
            update_list.append('GMAX')
            gmax_input = params.get_dict()['GMAX']
        elif self.ctx.check_dos:  # add only if doscheck is done
            updated_params = True
            update_list.append('GMAX')
            gmax_input = default_values.get('GMAX')
        # check if any mandatory keys are not set and set them with the default
        # values if missing in input parameters
        for key, value in default_values.items():
            if key not in update_list and key not in set_vals:
                self.report(f'INFO: setting {key} to default value {value}')
                kkr_para.set_value(key, value)

        # check if Fermi lavel should be set with input value
        if self.ctx.ef_set is not None:
            update_list.append('ef_set')

        # check if emin should be changed:
        if self.ctx.iter > 1 and (self.ctx.dos_check_fail_reason == 'EMIN too high'):
            # decrease emin by self.ctx.delta_e for DOS run only
            emin_old = self.ctx.emin
            emin_new = emin_old - self.ctx.delta_e
            self.ctx.dos_params_dict['emin'] = emin_new
            # change emin_new to internal Ry units
            emin_new = (emin_new - self.ctx.efermi) / eV2Ry
            update_list.append('EMIN')

        # store updated nodes (also used via last_params in kkr_scf_wc)
        if updated_params:
            # set values that are updated
            if 'RCLUSTZ' in update_list:
                kkr_para.set_value('RCLUSTZ', self.ctx.r_cls)
                self.report(f'INFO: setting RCLUSTZ to {self.ctx.r_cls}')
            if 'RMAX' in update_list:
                kkr_para.set_value('RMAX', rmax_input)  # pylint: disable=possibly-used-before-assignment
                self.report(f'INFO: setting RMAX to {rmax_input} (needed for DOS check with KKRcode)')
            if 'GMAX' in update_list:
                kkr_para.set_value('GMAX', gmax_input)  # pylint: disable=possibly-used-before-assignment
                self.report(f'INFO: setting GMAX to {gmax_input} (needed for DOS check with KKRcode)')
            if 'EMIN' in update_list:
                kkr_para.set_value('EMIN', emin_new)
                self.report(f'INFO: setting EMIN to {emin_new} (DOS check showed that EMIN is too high)')
            if 'ef_set' in update_list:
                kkr_para.set_value('EF_SET', self.ctx.ef_set)
                self.report(f'INFO: setting Fermi level of stating potential to {self.ctx.ef_set}')

            updatenode = orm.Dict(dict=kkr_para.get_dict())
            updatenode.description = f'changed values: {update_list}'
            if first_iter:
                updatenode.label = 'initial params from wf input'
                # used workfunction for provenance tracking if parameters have been changed
                params = update_params_wf(self.ctx.last_params, updatenode)
                self.ctx.last_params = params
            else:
                updatenode.label = f'updated params: {update_list}'
                # also keep track of last voronoi output if that has been used
                voro_out = self.ctx.voro_calc.outputs.output_parameters
                params = update_voro_input(self.ctx.last_params, updatenode, voro_out)
                self.ctx.last_params = params

        # run voronoi step
        options = {
            'queue_name': self.ctx.queue,
            'resources': self.ctx.resources,
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'custom_scheduler_commands': self.ctx.custom_scheduler_commands
        }

        if 'structure' in self.inputs:
            # normal mode starting from structure
            builder = get_inputs_voronoi(
                voronoicode,
                structure,
                options,
                label,
                description,
                params=params,
            )
        else:
            # parent_KKR mode
            builder = get_inputs_voronoi(
                voronoicode,
                None,
                options,
                label,
                description,
                params=params,
                parent_KKR=self.inputs.parent_KKR,
            )
        if 'startpot_overwrite' in self.inputs:
            builder.potential_overwrite = self.inputs.startpot_overwrite
        self.report(f'INFO: run voronoi step {self.ctx.iter}')
        future = self.submit(builder)

        # return remote_voro (passed to dos calculation as input)
        return ToContext(voro_calc=future)

    def check_voronoi(self):
        """Check voronoi output.

        return True/False if voronoi output is ok/problematic
        if output is problematic try to increase some parameters
        (e.g. cluster radius) and rerun up tp N_rerun_max times
        initializes with returning True
        """

        # check wether or not calculation was taken from cached node
        caching_info = f'INFO: cache_source of voro calc node: {self.ctx.voro_calc.get_cache_source()}'
        print(caching_info)
        self.report(caching_info)

        # do some checks with the voronoi output (finally sets self.ctx.voro_ok)
        self.ctx.voro_ok = True

        # TODO: print some info if voro_calc does not finish ok (need to find the reason for it)

        print('check_voro_calc', self.ctx.voro_calc, self.ctx.voro_calc.is_finished_ok)
        from pprint import pprint
        pprint(self.ctx.voro_calc.attributes)

        # check calculation state (calculation must be completed)
        if not self.ctx.voro_calc.is_finished_ok:
            self.report('ERROR: Voronoi calculation not in FINISHED state')
            self.ctx.voro_ok = False
            print('abort with', self.exit_codes.ERROR_VORONOI_FAILED)  # pylint: disable=no-member
            self.ctx.exit_code = self.exit_codes.ERROR_VORONOI_FAILED  # pylint: disable=no-member
            return False

        # check if parser returned some error
        voro_parser_errors = self.ctx.voro_calc.res.parser_errors
        if voro_parser_errors != []:
            self.report(f'ERROR: Voronoi Parser returned Error(s): {voro_parser_errors}')
            self.ctx.voro_ok = False
            self.ctx.exit_code = self.exit_codes.ERROR_VORONOI_PARSING_FAILED  # pylint: disable=no-member
            return False

        # check self.ctx.nclsmin condition
        clsinfo = self.ctx.voro_calc.res.cluster_info_group
        ncls = clsinfo.pop('number_of_clusters')

        nclsmin_last_calc = 1000
        for icls in range(len(clsinfo['cluster_info_atoms'])):
            tmp_ncls = clsinfo['cluster_info_atoms'][icls]['sites']
            if tmp_ncls < nclsmin_last_calc:
                nclsmin_last_calc = tmp_ncls
        self.report(f'INFO: number of atoms in smallest cluster: {nclsmin_last_calc}')

        if self.ctx.nclsmin > 0 and self.ctx.nclsmin > nclsmin_last_calc or ncls < 1:
            self.report(f'WARNING: minimal cluster smaller than threshold of {self.ctx.nclsmin}')
            self.ctx.voro_ok = False

        # check radii condition
        radii = self.ctx.voro_calc.res.radii_atoms_group
        r_ratio1 = radii[0]['rout_over_dist_nn']
        r_ratio2 = radii[0]['rmt0_over_rout']
        if r_ratio1 >= 100. or r_ratio2 >= 100.:
            self.report(f'ERROR: radii information inconsistent: Rout/dis_NN={r_ratio1}, RMT0/Rout={r_ratio2}')
            self.ctx.voro_ok = False
            self.ctx.exit_code = self.exit_codes.ERROR_VORONOI_INVALID_RADII  # pylint: disable=no-member
            return False

        # fix emin/emax
        # remember: efermi, emin and emax are in internal units (Ry) but delta_e is in eV!
        ret = self.ctx.voro_calc.outputs.retrieved
        if 'potential_overwrite' in self.ctx.voro_calc.inputs:
            potfile_overwrite = self.ctx.voro_calc.inputs.potential_overwrite
            with potfile_overwrite.open(potfile_overwrite.filename) as _file:
                self.ctx.efermi = get_ef_from_potfile(_file)
        else:
            potfile_name = VoronoiCalculation._OUT_POTENTIAL_voronoi
            if 'parent_KKR' in self.ctx.voro_calc.inputs:
                potfile_name = VoronoiCalculation._POTENTIAL_IN_OVERWRITE
            print(ret.list_object_names(), potfile_name)
            with ret.open(potfile_name) as _file:
                self.ctx.efermi = get_ef_from_potfile(_file)

        emin_dos = self.ctx.dos_params_dict['emin'] * eV2Ry + self.ctx.efermi
        if self.ctx.iter == 1:
            self.ctx.emin = self.ctx.voro_calc.res.emin
        emin_out = self.ctx.emin
        self.report(f'INFO: emin dos input: {emin_dos}, emin voronoi output: {emin_out}')
        if emin_out - self.ctx.delta_e * eV2Ry < emin_dos:
            emin_new = emin_out - self.ctx.delta_e * eV2Ry
            emin_new = (emin_new - self.ctx.efermi) / eV2Ry  # convert to relative eV units
            self.ctx.dos_params_dict['emin'] = emin_new
            self.report(f'INFO: set emin for dos automatically'
                        f' to {emin_out - self.ctx.delta_e * eV2Ry} Ry')

        emax = self.ctx.dos_params_dict['emax']
        self.report(f'INFO: emax dos input: {emax}, efermi voronoi output: {self.ctx.efermi}')
        if emax < self.ctx.delta_e:
            self.ctx.dos_params_dict['emax'] = self.ctx.delta_e
            self.report(
                f'INFO: self.ctx.efermi ({self.ctx.efermi} Ry) +'
                f' delta_e ({self.ctx.delta_e * eV2Ry} Ry) larger than'
                f' emax ({emax} Ry).'
                f' Setting automatically to {self.ctx.efermi + self.ctx.delta_e * eV2Ry} Ry'
            )

        #TODO implement other checks?

        self.report(f'INFO: Voronoi check finished with result: {self.ctx.voro_ok}')

        # finally return result of check
        return self.ctx.voro_ok

    def do_iteration_check(self):
        """Check if another iteration should be done."""
        if self.ctx.exit_code is not None:
            # break out of while loop if an error was caught
            return False
        if self.ctx.is_starting_iter:
            # initial iteration (at least one has to be done)
            # reset starting iter flag
            self.ctx.is_starting_iter = False
            return True
        if self.ctx.iter >= self.ctx.Nrerun:
            # check if maximal number of iterations is reached
            return False
        if self.ctx.voro_ok and self.ctx.doscheck_ok:
            # if both steps succeed we are done
            return False
        return True

    def get_dos(self):
        """Call dos sub workflow and pass the input and submit the calculation."""
        if self.ctx.check_dos:
            self.report(f'INFO: Doing DOS calculation in iteration {self.ctx.iter}')
            # take subset of input and prepare parameter node for dos workflow
            options_dict = {
                'queue_name': self.ctx.queue,
                'resources': self.ctx.resources,
                'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
                'withmpi': self.ctx.withmpi,
                'custom_scheduler_commands': self.ctx.custom_scheduler_commands
            }
            options_node = orm.Dict(dict=options_dict)
            options_node.label = 'options'
            wfdospara_node = orm.Dict(dict=self.ctx.dos_params_dict)
            wfdospara_node.label = 'DOS params'
            wfdospara_node.description = 'DOS parameters passed from kkr_startpot_wc input to DOS sub-workflow'

            code = self.inputs.kkr
            print('VORO_CALC:', self.ctx.voro_calc)
            print('VORO_CALC ok?', self.ctx.voro_calc.is_finished_ok)
            print('OUTGOINGS:', self.ctx.voro_calc.get_outgoing().all())
            remote = self.ctx.voro_calc.outputs.remote_folder
            wf_label = 'DOS calculation'
            wf_desc = 'subworkflow of a DOS calculation that perform a singe-shot KKR calc.'

            builder = kkr_dos_wc.get_builder()
            builder.metadata.description = wf_desc  # pylint: disable=no-member
            builder.metadata.label = wf_label  # pylint: disable=no-member
            builder.kkr = code
            builder.wf_parameters = wfdospara_node
            builder.options = options_node
            builder.remote_data = remote

            future = self.submit(builder)

            return ToContext(doscal=future)

    def check_dos(self):
        """Checks if dos of starting potential is ok."""
        dos_ok = True
        self.ctx.dos_check_fail_reason = None

        if self.ctx.check_dos:
            # check parser output
            doscal = self.ctx.doscal
            print('check doscal', doscal)
            from pprint import pprint
            pprint(doscal.attributes)
            try:
                dos_outdict = doscal.outputs.results_wf.get_dict()
                print(dos_outdict)
                if not dos_outdict['successful']:
                    self.report('ERROR: DOS workflow unsuccessful')
                    self.ctx.doscheck_ok = False
                    self.ctx.exit_code = self.exit_codes.ERROR_DOSRUN_FAILED  # pylint: disable=no-member

                if dos_outdict['list_of_errors'] != []:
                    self.report(f'ERROR: DOS wf output contains errors: {dos_outdict["list_of_errors"]}')
                    self.ctx.doscheck_ok = False
                    self.ctx.exit_code = self.exit_codes.ERROR_DOSRUN_FAILED  # pylint: disable=no-member
            except:
                self.ctx.doscheck_ok = False
                self.ctx.exit_code = self.exit_codes.ERROR_DOSRUN_FAILED  # pylint: disable=no-member

            # needed for checks
            emin = self.ctx.voro_calc.res.emin

            # check for negative DOS
            try:
                dosdata = doscal.outputs.dos_data
                natom = len(self.ctx.voro_calc.res.shapes)
                nspin = dos_outdict['nspin']

                ener = dosdata.get_x()[1]  # shape= natom*nspin, nept
                totdos = dosdata.get_y()[0][1]  # shape= natom*nspin, nept

                if len(ener) != nspin * natom:
                    self.report(
                        f'ERROR: DOS output shape does not fit nspin,'
                        f' natom information: len(energies)={len(ener)},'
                        f' natom={natom}, nspin={nspin}'
                    )
                    self.ctx.doscheck_ok = False
                    self.ctx.exit_code = self.exit_codes.ERROR_DOSRUN_FAILED  # pylint: disable=no-member

                # deal with snpin==1 or 2 cases and check negtive DOS
                for iatom in range(natom // nspin):
                    for ispin in range(nspin):
                        x, y = ener[iatom * nspin + ispin], totdos[iatom * nspin + ispin]
                        if nspin == 2 and ispin == 0:
                            y = -y
                        if y.min() < 0:
                            self.report(
                                f'INFO: negative DOS value found in'
                                f' (atom, spin)=({iatom},{ispin}) at'
                                f' iteration {self.ctx.iter}'
                            )
                            dos_ok = False
                            self.ctx.dos_check_fail_reason = 'DOS negative'

                # check starting EMIN
                dosdata_interpol = doscal.outputs.dos_data_interpol

                ener = dosdata_interpol.get_x()[1]  # shape= natom*nspin, nept
                totdos = dosdata_interpol.get_y()[0][1]  # shape= natom*nspin, nept [0] for total DOS
                Ry2eV = get_Ry2eV()

                for iatom in range(natom // nspin):
                    for ispin in range(nspin):
                        x, y = ener[iatom * nspin + ispin], totdos[iatom * nspin + ispin]
                        xrel = abs(x - (emin - self.ctx.efermi) * Ry2eV)
                        mask_emin = np.where(xrel == xrel.min())
                        ymin = abs(y[mask_emin])
                        if ymin > self.ctx.threshold_dos_zero:
                            self.report(f'INFO: DOS at emin not zero! {ymin}>{self.ctx.threshold_dos_zero}')
                            dos_ok = False
                            self.ctx.dos_check_fail_reason = 'EMIN too high'
            except AttributeError:
                dos_ok = False

            # TODO check for semi-core-states

            # TODO check rest of dos_output node if something seems important to check

        # finally set the value in context (needed in do_iteration_check)
        if dos_ok:
            self.ctx.doscheck_ok = True
        else:
            self.ctx.doscheck_ok = False

        return None

    def return_results(self):
        """Return the results of the dos calculations

        This should run through and produce output nodes even if everything failed,
        therefore it only uses results from context.
        """

        # finalchecks before write out
        if not self.ctx.voro_ok or not self.ctx.doscheck_ok:
            self.ctx.successful = False

        self.report('INFO: create vorostart results nodes.')

        # voronoi outputs
        try:
            voro_pk = self.ctx.voro_calc.pk
        except AttributeError:
            voro_pk = None
        try:
            voro_calc = self.ctx.voro_calc.outputs.output_parameters
        except NotExistent:
            self.report(f'ERROR: Results ParameterNode of voronoi (pk={voro_pk}) not found')
            voro_calc = None
        try:
            voro_remote = self.ctx.voro_calc.outputs.remote_folder
        except NotExistent:
            self.report(f'ERROR: RemoteFolderNode of voronoi (pk={voro_pk}) not found')
            voro_remote = None
        try:
            last_params = self.ctx.last_params
        except AttributeError:
            self.report(f'ERROR: Input ParameterNode of voronoi (pk={voro_pk}) not found')
            last_params = None

        # dos calculation outputs
        try:
            doscal = self.ctx.doscal.outputs.results_wf
        except AttributeError:
            self.report('WARNING: Results ParameterNode of DOS calc not found')
            doscal = None
        try:
            dosdata = self.ctx.doscal.outputs.dos_data
        except AttributeError:
            self.report('WARNING: DOS data of DOS calc not found')
            dosdata = None
        try:
            dosdata_interpol = self.ctx.doscal.outputs.dos_data_interpol
        except AttributeError:
            self.report('WARNING: interpolated DOS data of DOS calc not found')
            dosdata_interpol = None

        self.report(f'INFO: last_voro_calc={self.ctx.voro_calc}')
        self.report(f'INFO: voro_results={voro_calc}')
        self.report(f'INFO: voro_remote={voro_remote}')
        self.report(f'INFO: last_params={last_params}')
        try:
            self.report(f'INFO: last doscal={self.ctx.doscal}')
            self.report(f'INFO: doscal_results={doscal}')
            self.report(f'INFO: dosdata={dosdata}')
            self.report(f'INFO: dosdata_interpol={dosdata_interpol}')
        except:
            self.report('INFO: no doscal data')

        # collect output nodes
        link_nodes = {}

        if voro_calc is not None and voro_remote is not None and last_params is not None:
            link_nodes['last_voronoi_results'] = voro_calc
            link_nodes['last_voronoi_remote'] = voro_remote
            link_nodes['last_params_voronoi'] = last_params
        if doscal is not None and dosdata is not None and dosdata_interpol is not None:
            link_nodes['last_doscal_results'] = doscal
            link_nodes['last_doscal_dosdata'] = dosdata
            link_nodes['last_doscal_dosdata_interpol'] = dosdata_interpol

        # create dict to store results of workflow output
        res_node_dict = {}
        res_node_dict['workflow_name'] = self.__class__.__name__
        res_node_dict['workflow_version'] = self._workflowversion
        res_node_dict['successful'] = self.ctx.successful
        res_node_dict['list_of_errors'] = self.ctx.errors
        res_node_dict['withmpi'] = self.ctx.withmpi
        res_node_dict['resources'] = self.ctx.resources
        res_node_dict['max_wallclock_seconds'] = self.ctx.max_wallclock_seconds
        res_node_dict['queue_name'] = self.ctx.queue
        res_node_dict['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        res_node_dict['dos_params'] = self.ctx.dos_params_dict
        res_node_dict['description'] = self.ctx.description_wf
        res_node_dict['label'] = self.ctx.label_wf
        res_node_dict['last_rclustz'] = self.ctx.r_cls
        res_node_dict['min_num_atoms_in_cluster'] = self.ctx.nclsmin
        res_node_dict['factor_rcls_increase'] = self.ctx.fac_clsincrease
        res_node_dict['last_iteration'] = self.ctx.iter
        res_node_dict['max_iterations'] = self.ctx.Nrerun
        res_node_dict['last_voro_ok'] = self.ctx.voro_ok
        res_node_dict['last_dos_ok'] = self.ctx.doscheck_ok
        res_node_dict['starting_fermi_energy'] = self.ctx.efermi

        # create output Dict node and connect to the rest of the output nodes
        res_node = create_out_dict_node(orm.Dict(dict=res_node_dict), **link_nodes)
        res_node.label = 'vorostart_wc_results'
        res_node.description = ''

        # fill output_nodes dict with
        self.out('results_vorostart_wc', res_node)
        for link_label, node in link_nodes.items():
            self.out(link_label, node)

        self.report('INFO: done with kkr_startpot workflow!\n')

    def find_cluster_radius_alat(self):
        """Find an estimate for the cluster radius.

        The radius is chosen such that it comes close to having nclsmin atoms
        in the cluster.
        """

        # some settings
        ncls_target = self.ctx.nclsmin
        structure = self.ctx.structure

        # first find cluster radius in Ang. units
        r_cls_ang = None
        if self.ctx.nclsmin > 0:
            r_cls_ang = find_cluster_radius(structure, ncls_target)[0]
        else:
            if 'calc_parameters' in self.inputs:
                r_cls_ang = self.inputs.calc_parameters.get_dict().get('RCLUSTZ', None)
        if r_cls_ang is None:
            raise ValueError('nclsmin<=0 but RCLUSTZ not in calc_parameters!')

        # get lattic constant
        cell = np.array(structure.cell)
        alat = get_alat_from_bravais(cell, structure.pbc[2])

        # now check whether input parameter contain use_input_alat key
        # and overwrite alat with used input alat if given
        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
            use_input_alat = params.get_dict().get('use_input_alat', False)
            alat_input = params.get_dict().get('ALATBASIS', None)
            if use_input_alat and alat_input is not None:
                alat = alat_input

        # now get r_cls in proper alat units
        r_cls_alat = r_cls_ang / alat

        # return cluster radius
        return r_cls_alat

    def error_handler(self):
        """Capture errors raised in validate_input"""
        if self.ctx.exit_code is not None:
            return self.ctx.exit_code


@calcfunction
def update_voro_input(params_old, updatenode, voro_output):
    """Pseudo wf used to keep track of updated parameters in voronoi calculation.

    voro_output only enters as dummy argument for correct connection but logic using this value is done somewhere else.
    """
    dummy = voro_output
    # voro_output is only dummy input to draw connection in graph

    updatenode_dict = updatenode.get_dict()
    new_parameternode = update_params(params_old, nodename=None, nodedesc=None, **updatenode_dict)
    return new_parameternode



# Workflow for STM swf_parameterson around a magnetic impurity

from aiida.engine import WorkChain, ToContext, if_, calcfunction
from aiida.orm import Dict, RemoteData, Code, CalcJobNode, WorkChainNode, Float, Bool, XyData, SinglefileData, List
from aiida.orm import Group, load_group
from aiida.common import LinkType
from aiida_kkr.workflows.kkr_imp_dos import kkr_imp_dos_wc
from aiida_kkr.workflows.gf_writeout import kkr_flex_wc
from aiida_kkr.workflows.kkr_imp_sub import kkr_imp_sub_wc
from aiida_kkr.tools.find_parent import get_calc_from_remote
from aiida_kkr.tools.common_workfunctions import test_and_get_codenode

__copyright__ = (u'Copyright (c), 2024, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.5'
__contributors__ = (u'Raffaele Aliberti', u'David Antognini Silva', u'Philipp Rüßmann')
_VERBOSE_ = False


class kkr_STM_wc(WorkChain):
    """
    Workchain for the Simulation of a (superconducting) STM tip around a magnetic impurity.
    The workchain uses the (converged) impurity calculation of a host system and combines them
    with vacuum sites in positions

    inputs::
        :param options: (Dict), computer options
        :param tip_position: (Dict), specify the position of the STM tip
        :param impurity_info: (Dict), information about the impurity cluster
        :param host_calc: (RemoteData), information about the host structure of the sample
        :param wf_parameters: (Dict), parameters that are used to run the workflow
        :param kkr: (Code), KKR host code for the writing out kkrflex files
        :param kkrimp: (Code), KKR impurity code for the normal state impurity scf and BdG impurity DOS calculation
        :param gf_writeout.params_kkr_overwrite (Dict), overwrite parameters for the GF calculation
        :param kkr_imp_sub.params_overwrite (Dict), overwrite parameters for the impurity calculation

     returns::

        :return workflow_info: (Dict), Information of workflow results
                            like success, last result node, list with convergence behavior
        :return STM_dos_data: (XYData), Returns the plot of the lmDOS of the calculation
        :retrun STM_lmdos_data: (XYData), Returns the interpolated lmDOS of the calculation"""

    # TO DO: Add the initialize step.
    # TO DO: Add a better creation of the impurity cluster.
    # TO DO: Add check that between the ilayer and the actual number of layers in the structure.
    # TO DO: Don't run the clustering step if kkrflexfiles are given: It's redundant and it can lead to errors

    _wf_version = __version__
    _wf_label = 'STM_wc'
    _wf_description = 'Workflow for simulating an STM measurement'

    _options_default = {
        'queue_name': '',  # Queue name to submit jobs too
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 48,
            'num_cores_per_mpiproc': 1
        },  # resources to allocate for the job
        'max_wallclock_seconds': 3600 * 2,  # walltime after which the job gets killed (gets parsed to KKR)}
        'custom_scheduler_commands': '',  # some additional scheduler commands
        'withmpi': True
    }  # execute KKR with mpi or without

    _wf_default = {
        'jij_run': False,  # calculate Jij's energy resolved
        'lmdos':
        True,  # calculate also (l,m) or only l-resolved DOS, for the STM wf this is alwyas set on True as default
        'retrieve_kkrflex': False,  # retrieve kkrflex files to repository or leave on remote computer only
    }
    # add defaults of dos_params since they are passed onto that workflow
    _wf_default['dos_params'] = kkr_imp_dos_wc.get_wf_defaults()['dos_params']

    # return default values (helpful for users)
    @classmethod
    def get_wf_defaults(cls, silent=False):
        """Print and return _wf_default dictionary.

        Can be used to easily create set of wf_parameters.
        returns _wf_default, _options_default
        """
        if not silent:
            print(f'Version of workflow: {cls._wf_version}')
        return cls._wf_default

    @classmethod
    def define(cls, spec):
        """
        Layout of the workflow, defines the input nodes and the outline of the workchain
        """
        super(kkr_STM_wc, cls).define(spec)

        spec.input('kkr', valid_type=Code, required=False, help='KKRhost code, needed if gf_dos_remote is not given.')
        spec.input('kkrimp', valid_type=Code, required=True, help='KKRimp code, always needed.')
        spec.input(
            'options',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._options_default),
            help='Computer options (resources, quene name, etc.).'
        )
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._wf_default),
            help='Workflow parameter (see `kkr_dos_wc.get_wf_defaults()`).'
        )
        spec.input(
            'tip_position',
            valid_type=Dict,
            required=False,
            # Find a way to create an area of this position.
            default=lambda: Dict({
                'ilayer': 0,
                'nx': 0,
                'ny': 0
            }),
            # In the previous line we have set to study the first layer
            # nx is the number of (symmetric) steps that we take in the x-direction starting from the impurity
            # ny is the number of (symmetric) steps that we take in the y-direction starting from the impurity
            # (0,0) correspond to calculate the DOS only on the impurity site
            help=
            'How many sites will be scanned in the da and db direction (Bravais Lattice). And the layer that is being scanned.'
        )
        spec.input(
            'imp_info',
            valid_type=Dict,
            required=True,
            help='Information of the impurity like position in the unit cell, screening cluster, atom type.'
        )
        spec.input(
            'host_remote',
            valid_type=RemoteData,
            required=True,
            help='Remote Data containing the remote folder from the outputs of the host calculation',
        )
        spec.input(
            'imp_potential_node',
            valid_type=SinglefileData,
            required=True,
            help='Impurity potential node',
        )
        spec.input(
            'kkrflex_files',
            valid_type=RemoteData,
            required=False,
            help='with this input we can directly load the gf_dos files without calculating them'
        )

        # Here we expose the inputs for the GF calculations step.
        # One parameter which is crucial is the NSHELD, which determines the impurity cluster radius.
        spec.expose_inputs(kkr_flex_wc, namespace='gf_writeout', include=('params_kkr_overwrite', 'options'))

        # Here we expose the BdG calculations from the kkr_imp_dos_wc
        spec.expose_inputs(kkr_imp_sub_wc, namespace='BdG', include=('params_overwrite'))
        spec.expose_inputs(kkr_imp_sub_wc, include=('initial_noco_angles', 'rimpshift'))

        # Specify the possible outputs
        spec.output('tip_position', valid_type=Dict)
        spec.output('STM_dos_data', valid_type=XyData, required=True)
        spec.output('STM_dos_data_lmdos', valid_type=XyData, required=True)
        #spec.output('workflow_info', valid_type=Dict)
        spec.output('kkrflexfiles', valid_type=RemoteData)
        spec.output('combined_imp_info', valid_type=Dict)
        spec.output('combined_imp_potential', valid_type=SinglefileData)

        # Define all possible error messages

        spec.exit_code(100, 'ERROR_STM_POSITION_NOT_VALID', 'The position provided for the STM probe are incorrect')
        spec.exit_code(101, 'ERROR_IMP_INFO_NOT_CORRECT', 'The node provided for the impurity info is not valid')
        spec.exit_code(102, 'ERROR_NO_IMP_POT_SFD', 'No impurity node has been given in the intput')
        spec.exit_code(103, 'ERROR_NO_IMPURITY_INFO', 'No impurity info has been given in the input')
        spec.exit_code(
            104, 'ERROR_NO_DATA_FOR_THE_GF_STEP', """Neither the kkrflex files nor the KKR builder have been given.
Please provide already converged kkrflex files, or the kkr builder to evaluate them"""
        )
        spec.exit_code(201, 'ERROR_IMP_SUB_WORKFLOW_FAILURE', 'A step in the kkr_imp_dos workflow has failed')

        spec.outline(
            # For initializing workflow
            cls.start,
            # We first aggregate all the impurity data
            # The gf is then used to evaluate the STM lmdos
            cls.STM_lmdos_run,
            cls.results
        )

    def combine_imp_info(self, host_structure, impurity_to_combine, da, db):
        from aiida_kkr.tools.tools_STM_scan import get_imp_info_add_position
        import numpy as np  # TO DO: optimize this call, only need append from numpy
        """
        Here we want to combine the impurity information and the host information
        """
        tip_position = self.get_tip_position_dict(da, db)
        imp_info = self.inputs.imp_info  #(impurity to combine)

        combined_imp_info = get_imp_info_add_position(Dict(tip_position), host_structure, imp_info)

        # If the position already exists we simply return the old dictionary
        if combined_imp_info is None:
            return impurity_to_combine

        # Add check to see if imp_cls is there
        if 'imp_cls' in impurity_to_combine:

            for key in impurity_to_combine.keys():
                if key == 'Zimp':
                    impurity_to_combine[key].append(combined_imp_info[key][-1])
                else:
                    impurity_to_combine[key] = np.append(impurity_to_combine[key], [combined_imp_info[key][-1]], axis=0)

            # Convert to an AiiDA Dictionary
            new_combined_imp_info = impurity_to_combine

        else:

            new_combined_imp_info = combined_imp_info

        return new_combined_imp_info

    def combine_potentials(self, host_calc, node_to_combine, da, db):
        from aiida_kkr.tools.tools_STM_scan import create_combined_potential_node
        """
        Here we create a combined potential node from the host potential (no impurity)
        and from the impurity potential
        """

        tip_position = self.get_tip_position_dict(da, db)
        combined_potential_node = create_combined_potential_node(tip_position, host_calc, node_to_combine)
        return combined_potential_node

    def get_tip_position_dict(self, da, db):
        # Since the objects in AiiDA are immutable we have to create a new dictionary
        # and then convert it to the right AiiDA type
        tip_position = {}
        tip_position['ilayer'] = self.inputs.tip_position['ilayer']
        tip_position['da'] = da
        tip_position['db'] = db
        return tip_position

    def start(self):
        """
        Initialise context and some parameters
        """

        self.report(f'INFO: started STM workflow version {self._wf_version}')
        if _VERBOSE_:
            self.report(f'inputs: {self.inputs}')

        # Input both wf and options parameters
        # We check if the inputs for the wc where given
        # Otherwise we assign the default values
        if 'options' in self.inputs:
            options_dict = self.inputs.options.get_dict()
            # empty dictionary evaluate as false to Python
            if options_dict == {}:
                options_dict = self._options_default
                self.report('INFO: using default options parameters')

        if 'wf_parameters' in self.inputs:
            wf_param_dict = self.inputs.wf_parameters.get_dict()
            # empty dictionary evaluate as false to Python
            if wf_param_dict == {}:
                options_dict = self._wf_default
                self.report('INFO: usign defalut wf parameters')

        # In this section we assign the computational resources to the builder
        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands', self._options_default['custom_scheduler_commands']
        )
        self.ctx.options_params_dict = Dict({
            'withmpi': self.ctx.withmpi,
            'resources': self.ctx.resources,
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'queue_name': self.ctx.queue,
            'custom_scheduler_commands': self.ctx.custom_scheduler_commands
        })

        # Set workflow parameters for the KKR imputrity calculations
        # This part is really important, this should always be set to True for an STM calculation
        self.ctx.lmdos = wf_param_dict.get('lmdos', self._wf_default['lmdos'])

        self.ctx.retrieve_kkrflex = wf_param_dict.get('retrieve_kkrflex', self._wf_default['retrieve_kkrflex'])

        self.ctx.dos_params_dict = wf_param_dict.get('dos_params', self._wf_default['dos_params'])

        # fill missing key, value pairs with defaults
        for k, v in self._wf_default['dos_params'].items():
            if k not in self.ctx.dos_params_dict.keys():
                self.ctx.dos_params_dict[k] = v

        # set workflow parameters for the KKR impurity calculation
        self.ctx.jij_run = wf_param_dict.get('jij_run', self._wf_default['jij_run'])

        # set workflow label and description
        self.ctx.description_wf = self.inputs.get('description', self._wf_description)
        self.ctx.label_wf = self.inputs.get('label', self._wf_label)

        if _VERBOSE_:
            message = f"""
                        INFO: use the following parameter:
                        withmpi: {self.ctx.withmpi}
                        Resources: {self.ctx.resources}
                        Walltime (s): {self.ctx.max_wallclock_seconds}
                        queue name: {self.ctx.queue}
                        scheduler command: {self.ctx.custom_scheduler_commands}
                        description: {self.ctx.description_wf}
                        label: {self.ctx.label_wf}
                      """
            self.report(message)

    def validate_input(self):
        """Check if inputs are valid"""
        inputs = self.inputs

        if not 'imp_potential_node' in inputs:
            return self.exit_codes.ERROR_NO_IMP_POT_SFD  # pylint: disable=no-member

        if not 'imp_info' in inputs:
            return self.exit_codes.ERROR_NO_IMP_INFO  # pylint: disable=no-member

        if not 'kkrflex_files' and 'kkr' in inputs:
            return self.exit_codes.ERROR_NO_DATA_FOR_THE_GF_STEP  # pylint: disable=no-member

    def impurity_cluster_evaluation(self):
        """
        Create the combined impurity cluster and potential for the impurity region
        used in self-consistency + the additional scanning sites.
        """
        from aiida_kkr.tools import find_parent_structure
        from aiida_kkr.tools.imp_cluster_tools import pos_exists_already
        from aiida_kkr.tools.tools_STM_scan import get_imp_cls_add, convert_to_imp_cls, offset_clust2

        if _VERBOSE_:
            from time import time
            # measure time at start
            t_start = time()

        # Here we create an impurity cluster that has inside all the positions on which the STM will scan

        impurity_info = self.inputs.imp_info  # for the first step we combine the impurity info from the input
        imp_potential_node = self.inputs.imp_potential_node  # for the first step we combine the impurity node from the input

        host_remote = self.inputs.host_remote
        host_calc = host_remote.get_incoming(node_class=CalcJobNode).first().node
        host_structure = find_parent_structure(host_remote)

        # now find all the positions we need to scan
        coeff = self.get_scanning_positions(host_remote)

        if _VERBOSE_:
            # timing counters
            t_imp_info, t_pot = 0., 0.

        _, imp_clust = convert_to_imp_cls(host_structure, impurity_info)

        # construct impurity potential and imp_info for the impurity cluster + scanning area

        # Check if the impuirty cluster already exists, if so create a new entity that can be modified
        if 'imp_cls' in impurity_info:
            impurity_info_aux = impurity_info.clone()
            imp_potential_node_aux = imp_potential_node.clone()
        else:  # Otherwise use the one from the inputs
            impurity_info_aux = impurity_info
            imp_potential_node_aux = imp_potential_node

        # Case in which we don't pass any element to embed in the impurity cluster, it
        # uses the impurity files given in the inputs.
        if len(coeff) == 0:
            return impurity_info, imp_potential_node

        for element in coeff:

            if _VERBOSE_:
                t0 = time()

            # Check if the position is already in the cluster
            # for this we need to first get the position
            tmp_pos = self.get_tip_position_dict(element[0], element[1])
            _, tmp_clust = get_imp_cls_add(host_structure, tmp_pos)
            clust_offset = offset_clust2(imp_clust, tmp_clust, host_structure, Dict(tmp_pos))

            if _VERBOSE_:
                t_cluster_offset += time() - t0

            if pos_exists_already(imp_clust[:, :3], clust_offset[:, :3])[0]:
                message = f'INFO: The position {tmp_pos} is already present in the system, skipping it'
                self.report(message)
                continue  # If the position exists already skip the embedding
            else:
                # Aggregation of the impurity potential
                tmp_imp_info = self.combine_imp_info(host_structure, impurity_info_aux, element[0], element[1])
                impurity_info_aux = tmp_imp_info

                if _VERBOSE_:
                    self.report('imp info has been embedded')
                    t_imp_info += time() - t0
                    t0 = time()

                # Aggregation the impurity nodes
                tmp_imp_pot = self.combine_potentials(host_calc, imp_potential_node_aux, element[0], element[1])
                imp_potential_node_aux = tmp_imp_pot

                if _VERBOSE_:
                    self.report('imp potential has been added')
                    t_pot += time() - t0

        if _VERBOSE_:
            # report elapsed time for cluster generation
            self.report(
                f'time for cluster generation (s): {time()-t_start}, cluster generation={t_cluster_offset}, imp_info={t_imp_info}, pot={t_pot}'
            )

        return impurity_info_aux, imp_potential_node_aux

    def get_scanning_positions(self, host_remote):
        """
        Extract scanning positions either from input 'scan_positions' or from 'nx', 'ny' + symmetry analysis

        If  'scan_positions' is found in 'tip_position' input dict we use these positions which should
        be 2D array of integers with the positions in units of the structure's in-plane Bravais matrix.

        Otherwise we use the 'nx', 'ny' input to define a scanning region where an automated symmetry
        analysis is done to reduce the scanning area to the irreducible part.
        """
        # TO DO update this tool to get scanning positions even in the presence of more than one impurity
        from aiida_kkr.tools import tools_STM_scan

        generate_scan_positions = True
        if 'scan_positions' in self.inputs.tip_position:
            coeff = self.inputs.tip_position['scan_positions']
            if coeff is not None:
                # check if coefficients exists and are valid
                # TODO: improve the validity check
                generate_scan_positions = False

                if _VERBOSE_:
                    if generate_scan_positions:
                        self.report('The scanning positions have been given by the user')
                    else:
                        self.repotr('The scanning positions were automatically generated')

        if generate_scan_positions:

            # Information of the host structure
            struc_info, symm_matrices = tools_STM_scan.STM_pathfinder(host_remote)

            # We now want to iterate over several in-plane positions.
            # These are the number of vectors in which we want to move the STM tip.
            nx = self.inputs.tip_position['nx']
            ny = self.inputs.tip_position['ny']

            # Path creation step. (The the identity operator is present, but will be excluded)
            unused_pos, used_pos = tools_STM_scan.lattice_generation(
                symm_matrices, struc_info['plane_vectors'], 0, 0, nx, ny
            )

            # Since the combine tools use the element already in the units of da and db, we use a helper function
            # to have the indices of the linear combination of the used position vectors in the base of the Bravais lattice.
            coeff = tools_STM_scan.find_linear_combination_coefficients(struc_info['plane_vectors'], used_pos)

        return coeff

    def STM_lmdos_run(self):
        """In this part of the worflow we want to simulate the lmdos which a STM is able to measure """

        # First we would like to distinguish between an impurity dos and a normal state calculation
        builder = kkr_imp_dos_wc.get_builder()

        # Code loading
        builder.kkrimp = self.inputs.kkrimp  # needed for the kkr_imp_dos_wc

        # Builder options
        builder.options = self.ctx.options_params_dict

        # Check if the kkrflex files are already given in the outputs
        if 'kkrflex_files' in self.inputs:
            builder.gf_dos_remote = self.inputs.kkrflex_files
            message = f"""Remote host function is given in the outputs from the node: {self.inputs.kkrflex_files}. Please also make sure of
                          using the right impurity potentials from the already converged calculation."""
            self.report(message)
        else:
            builder.kkr = self.inputs.kkr  # needed to evaluate the kkr_flex files in the DOS step

        # NSHELD is the parameter that controls the radius in the impurity cluster.
        # The bigger the scanning position, the greater it must be set.
        if 'gf_writeout' in self.inputs:
            if 'params_kkr_overwrite' in self.inputs.gf_writeout:
                builder.gf_writeout.params_kkr_overwrite = self.inputs.gf_writeout.params_kkr_overwrite  # pylint: disable=no-member
            if 'options' in self.inputs.gf_writeout:
                builder.gf_writeout.options = self.inputs.gf_writeout.options  # pylint: disable=no-member
        else:
            # This is a big value of NSHELD to make sure that most calculations work
            builder.gf_writeout.params_kkr_overwrite = Dict(dict={'NSHELD': 1500})  # pylint: disable=no-member

        # Update the BdG parameters if they are inserted in the workflow
        if 'BdG' in self.inputs:
            if 'params_overwrite' in self.inputs.BdG:
                builder.BdG.params_overwrite = self.inputs.BdG.params_overwrite  # pylint: disable=no-member

        self.ctx.kkrimp_params_dict = Dict(
            dict={
                'nsteps': 1,  # redundant because this is already set inside the kkr_imp_dos workchain?!
                'kkr_runmax': 1,  # redundant because this is already set inside the kkr_imp_dos workchain?!
                'dos_run': True,  # redundant because this is already set inside the kkr_imp_dos workchain?!
                'retrieve_kkrflex': self.ctx.retrieve_kkrflex,
                'lmdos': self.ctx.lmdos,
                'jij_run': self.ctx.jij_run,
                'dos_params': self.ctx.dos_params_dict
            }
        )

        # We want to set the energy to the Fermi level
        if 'emin' not in self.ctx.dos_params_dict:
            self.ctx.kkrimp_params_dict['dos_params']['emin'] = 0 - 0.005
        if 'emax' not in self.ctx.dos_params_dict:
            self.ctx.kkrimp_params_dict['dos_params']['emax'] = 0 + 0.005

        # Finally we overwrite the number of energy points to 1
        # This is because we want many epoints around the impurity position
        if 'nepts' not in self.ctx.dos_params_dict:
            self.ctx.kkrimp_params_dict['dos_params'][
                'nepts'] = 7  # Here 7 because of the interpolated files that aren't generated

        builder.wf_parameters = self.ctx.kkrimp_params_dict
        # Host remote files that will be used for the actual plot step.
        builder.host_remote = self.inputs.host_remote

        # Here we create the impurity cluster for the STM scanning tool

        #if 'Rcut' in self.inputs.imp_info.get_dict():
        #    # If the data doesn't come from a previous calculation we create it
        #    impurity_info, imp_pot_sfd = self.impurity_cluster_evaluation()
        #else:
        #    impurity_info = self.inputs.imp_info
        #    imp_pot_sfd = self.inputs.imp_potential_node

        impurity_info, imp_pot_sfd = self.impurity_cluster_evaluation()

        # With this we make sure that the actual number of angles is the same as the number of embedded impurity
        if 'initial_noco_angles' in self.inputs:
            inital_noco_angles_aux = self.inputs.initial_noco_angles.clone()
            for imp in impurity_info.get_dict()['Zimp']:
                if imp == 0:
                    inital_noco_angles_aux.get_dict()['phi'].append(0.0)
                    inital_noco_angles_aux.get_dict()['theta'].append(0.0)
                    inital_noco_angles_aux.get_dict()['fix_dir'].append(1)

        # impurity info for the workflow
        builder.impurity_info = impurity_info
        builder.imp_pot_sfd = imp_pot_sfd

        # submit calculation
        calc = self.submit(builder)

        # print report
        message = f"""INFO: running DOS step for an STM measurement (pk: {calc.pk}) at position (ilayer: {self.inputs.tip_position['ilayer']})"""
        if 'params_overwrite' in self.inputs.BdG:
            if self.inputs.BdG.params_overwrite:
                message += f'\nINFO: runnig DOS step (pk: {calc.pk}) BdG is present'
        self.report(message)

        # Save the calculated impurity cluster and impurity info in the context

        self.ctx.impurity_info = impurity_info
        self.ctx.imp_pot_sfd = imp_pot_sfd

        return ToContext(STM_data=calc)

    def results(self):
        """Collect results and return output nodes"""

        if not self.ctx.STM_data.is_finished_ok:
            self.report('ERROR: sub workflow for STM calculation failed')
            return self.exit_codes.ERROR_IMP_SUB_WORKFLOW_FAILURE  # pylint: disable=no-member
        else:
            # Declaring the output
            self.out('STM_dos_data', self.ctx.STM_data.outputs.dos_data)
            self.out('STM_dos_data_lmdos', self.ctx.STM_data.outputs.dos_data_lm)
            self.out('tip_position', self.inputs.tip_position)
            if 'gf_dos_remote' in self.ctx.STM_data.outputs:
                self.out('kkrflexfiles', self.ctx.STM_data.outputs.gf_dos_remote)
            self.out('combined_imp_info', self.ctx.impurity_info)
            self.out('combined_imp_potential', self.ctx.imp_pot_sfd)

        self.report('INFO: created output nodes for KKR STM workflow.')

        self.report(
            '\n'
            '|------------------------------------------------------------------------------------------------------------------|\n'
            '|-----------------------------------------| Done with the STM workflow! |------------------------------------------|\n'
            '|------------------------------------------------------------------------------------------------------------------|'
        )



# -*- coding: utf-8 -*-
"""
Input plug-in for a KKRnano calculation.
"""

import numpy as np
from masci_tools.io.common_functions import get_Ang2aBohr
from aiida.orm import CalcJobNode, Dict, Bool, Float, RemoteData, StructureData
from aiida.engine import CalcJob
from aiida.common import NotExistent
from aiida.common.datastructures import (CalcInfo, CodeInfo)
from aiida.common.exceptions import InputValidationError, UniquenessError
from aiida_kkr.tools.find_parent import get_remote, get_parent
from aiida_kkr.data.strucwithpot import StrucWithPotData

__copyright__ = (u'Copyright (c), 2021, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.0.2'
__contributors__ = ('Markus Struckmann', 'Philipp Rüßmann')


class KKRnanoCalculation(CalcJob):
    """
    AiiDA calculation plugin for a KKRnano calculation
    """

    ####################
    # File names etc.
    ####################
    # calculation plugin version
    _CALCULATION_PLUGIN_VERSION = __version__
    # Default input and output files
    _DEFAULT_INPUT_FILE = 'input.conf'  # will be shown with inputcat
    _DEFAULT_EFERMI_FILE = 'EFERMI'
    _DEFAULT_NOCO_INPUT_FILE = 'nonco_angle.dat'
    _DEFAULT_OUTPUT_FILE = 'out'  #'shell output will be shown with outputcat
    _CONVERT_OUTPUT_FILE = 'convert_out'
    _DEFAULT_OUTPUT_PREP_FILE = 'output.0.txt'
    _DEFAULT_NOCO_OUTPUT_FILE = 'nonco_angle_out.dat'
    # template.product entry point defined in setup.json
    _DEFAULT_PARSER = 'kkr.kkrnanoparser'
    # File names
    _SHAPEFUN = 'shapefun'
    _POTENTIAL = 'potential'
    _OUT_POTENTIAL = 'out_potential'
    _RBASIS = 'rbasis.xyz'





    _DEFAULT_KKRNANO_PARA = {
      'bzdivide': {'value': [10,10,10],'required': True,'description': 'number of k-points in each direction'},
      'emin': {'value': -1.2, 'unit':'Rydberg', 'required': True, 'description':  'lower energy of contour'},
      'emax': {'value': 1.0 , 'unit':'Rydberg',  'required': True,\
               'description':  'upper energy of contour (relevant only for DOS calculation)'},
      'npnt1': {'value': 3, 'unit':'Rydberg',  'required': True,\
                'description': 'number of points starting at emin, parallel to imaginary axis'},
      'npnt2': {'value':  20,  'required': True,\
                'description': 'number of points parallel to real axis starting from emin + imag part.'},
      'npnt3': {'value':  3,  'required': True,\
                'description': 'number of points parallel to real axis in interval (E_F - 30*k*T + imag, E_F + imag)'},
      'npol': {'value':  7,  'required': True,\
               'description': 'Number of Matsubara poles, npol=0 triggers DOS calculation'} ,

      'scfsteps': {'value':  1,'required': True, 'description': 'Number of scf steps'},
      'imix': {'value':  6,'required': True,\
               'description': "mixing method: imix = 0 -> straight mixing; imix = 1 -> straight mixing;\
               imix = 4 -> Broyden's 2nd method; imix = 5 -> gen. Anderson mixing;\
               imix = 6 -> Broyden's 2nd method with support for >1 atom per process"
              },

      'mixing': {'value':  0.01, 'required': True,\
                 'description': 'straight mixing parameter'},
      'rmax': {'value':  8.0, 'required': True,\
               'description': 'Ewald sum cutoff in real space in units of lattice constants'} ,
      'gmax': {'value':  48.0, 'required': True,\
               'description': 'Ewald sum cutoff in reciprocal space in units of 2*Pi/alat'},
      'nsra': {'value':  2, 'required': True, 'description': '1=non-scalar-relativistic 2=scalar-relativistic'},
      'kte': {'value':  1, 'required': True,\
              'description': '1=calculate energies, -1 = total energy only, less I/O'} ,
      'rclust_voronoi': {'value':  2.00, 'required': True,\
                         'description': 'radius of cluster used for Voronoi (Should be irrelevant for use with aiida)'}#,
      #"soc":{"value": True}
    }

    @classmethod
    def define(cls, spec):
        """
        define internals and inputs / outputs of calculation
        """
        # reuse base class (i.e. CalcJob) functions
        super(KKRnanoCalculation, cls).define(spec)
        # now define input files and parser
        spec.inputs['metadata']['options']['parser_name'].default = cls._DEFAULT_PARSER
        spec.inputs['metadata']['options']['input_filename'].default = cls._DEFAULT_INPUT_FILE
        spec.inputs['metadata']['options']['output_filename'].default = cls._DEFAULT_OUTPUT_FILE
        #spec.inputs['metadata']['options']['output_prep_filename'].default = cls._DEFAULT_OUTPUT_PREP_FILE ->does not work

        # define input nodes (optional ones have required=False)
        spec.input(
            'parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._DEFAULT_KKRNANO_PARA),
            help='Dict node that specifies the input parameters for KKRnano (k-point density etc.)'
        )
        spec.input(
            'nocoangles',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict={}),
            help='Dict node that specifies the starting angles for non-colinear calculations\
                   (only needed in conjunction with non-colinear calculations, i. e. KORBIT=1\
                   (which is also necessary for SOC calculations!))'
        )
        spec.input(
            'parent_folder',
            valid_type=RemoteData,
            required=False,
            help='Use a node that specifies a parent KKRnano or voronoi calculation'
        )

        spec.input(
            'convert',
            valid_type=Bool,
            required=False,
            default=lambda: Bool(False),
            help='Activate to use together with set up convert code in order to retrieve potential files.'
        )

        spec.input(
            'passed_lattice_param_angs',
            valid_type=Float,
            required=False,
            default=lambda: Float(-10000.0),
            help='Use a prespecified lattice constant in Angstrom as input for KKRnano, i. e. in the input.conf file. \
                   Default is the length of the longest Bravais vector in the structure object used for the voronoi calculation. \
                   This can be useful in the context of treating supercells.'
        )

        spec.input('strucwithpot', valid_type=StrucWithPotData, required=False)

        #spec.input('structure', valid_type=StructureData, required=False, default:)
        # define outputs
        spec.output('output_parameters', valid_type=Dict, required=True, help='results of the calculation')
        spec.default_output_node = 'output_parameters'

        # define exit codes, also used in parser
        spec.exit_code(301, 'ERROR_NO_OUTPUT_FILE', message='KKRnano output file not found')
        spec.exit_code(302, 'ERROR_PARSING_FAILED', message='KKRnano parser retuned an error')

    def prepare_for_submission(self, tempfolder):
        """Create the input files from the input nodes passed to this instance of the `CalcJob`.

        :param tempfolder: an `aiida.common.folders.Folder` to temporarily write files on disk
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        #prepare inputs
        parameters = self.inputs.parameters.get_dict()
        #passed_lattice_const = self.inputs.passed_lattice_param_angs.value
        nonco_angles = self.inputs.nocoangles.get_dict()
        #parent_calc=self.inputs.parent_folder#.get_incoming(node_class=CalcJobNode).first().node
        #parent_calc_calc_node=parent_calc.get_incoming(node_class=CalcJobNode).first().node
        #structure = self._find_parent_struc_from_voro_or_stwpd(parent_calc_calc_node)[0].get_pymatgen_structure()
        #parent_outfolder = parent_calc_calc_node.outputs.retrieved
        code = self.inputs.code

        #determine number of MPI processes
        try:
            num_mpi_procs = self.metadata.options.resources['tot_num_mpiprocs']
        except:
            try:
                num_mpi_procs=self.metadata.options.resources['num_machines']*\
                self.metadata.options.resources['num_mpiprocs_per_machine']
            except:
                raise InputValidationError(
                    "The total number of MPI processes could not be determined. In case of doubt: Specify `builder.metadata.options.resources['tot_num_mpiprocs']=?` Do not forget number of machines"
                )

        #Check if convert mode has been activated
        convert = self.inputs.convert.value

        #StrucWithPot object as starting point -> contains passed_lattice_constant, shapefun, potential, and structure
        if hasattr(self.inputs, 'strucwithpot') and not hasattr(self.inputs, 'parent_folder'):
            use_strucwithpot = True
            strucwithpot = self.inputs.strucwithpot
            structure = strucwithpot.structure.get_pymatgen_structure()

            if self.inputs.passed_lattice_param_angs.value == -10000.0:  #check if default value is used
                if hasattr(strucwithpot, 'specified_lattice_constant'):
                    passed_lattice_const = strucwithpot.specified_lattice_constant.value
                else:
                    passed_lattice_const = self.inputs.passed_lattice_param_angs.value
            else:
                passed_lattice_const = self.inputs.passed_lattice_param_angs.value

            parent_outfolder_uuid = None  #use dummy so that copylists are still available
        elif not hasattr(self.inputs, 'strucwithpot') and hasattr(self.inputs, 'parent_folder'):
            use_strucwithpot = False
            passed_lattice_const = self.inputs.passed_lattice_param_angs.value
            parent_calc = self.inputs.parent_folder  #.get_incoming(node_class=CalcJobNode).first().node
            parent_calc_calc_node = parent_calc.get_incoming(node_class=CalcJobNode).first().node

            #Find structure unless convert mode is activated, as for that no structure is needed
            if not convert:
                structure = self.find_parent_struc_from_voro_or_stwpd(parent_calc_calc_node)[0].get_pymatgen_structure()
            parent_outfolder = parent_calc_calc_node.outputs.retrieved

            parent_outfolder_uuid = parent_outfolder.uuid

        else:
            raise InputValidationError(
                'Either `strucwithpot` or a `parent_folder` has to be provided. If necessary remove one of the inputs.'
            )

        print('passed lattice constant=', passed_lattice_const)

        # Local copy list for continuing KKRnano calculations (Note: parent_outfolder_uuid is None if use_strucwithpot)
        # if necessary, change some names in order to start from a preconverged calculation
        # (cf. jukkr/source/KKRnano/scripts/prepare.sh)
        local_copy_list_for_continued = [(parent_outfolder_uuid, 'bin.vpotnew', 'bin.vpotnew.0'),
                                         (parent_outfolder_uuid, 'bin.vpotnew.idx', 'bin.vpotnew.0.idx'),
                                         (parent_outfolder_uuid, 'bin.meshes', 'bin.meshes.0'),
                                         (parent_outfolder_uuid, 'bin.meshes.idx', 'bin.meshes.0.idx'),
                                         (parent_outfolder_uuid, 'bin.energy_mesh', 'bin.energy_mesh.0'),
                                         (parent_outfolder_uuid, 'bin.atoms', 'bin.atoms')]

        #(parent_outfolder_uuid,"nonco_angle_out.dat","bin.atoms"  )]
        #TODO: Parse noco_angle_out.dat and write a new file from that

        #Convert mode is not available for call using strucwithpot
        if use_strucwithpot and convert:
            raise InputValidationError('Convert mode cannot be used for call using a strucwithpot object.')
        if convert:
            if not parent_calc_calc_node.process_label == 'KKRnanoCalculation':
                raise InputValidationError(
                    'A convert process can only be started from a KKRnano calculation. Check also that convert code is used!'
                )

            #mark files from previous calc as such (Note: parent_outfolder_uuid is None if use_strucwithpot)
            local_copy_list_for_continued = [
                (parent_outfolder_uuid, 'bin.vpotnew', 'bin.vpotnew'),
                (parent_outfolder_uuid, 'bin.vpotnew.idx', 'bin.vpotnew.idx'),
                (parent_outfolder_uuid, 'bin.meshes', 'bin.meshes'),
                (parent_outfolder_uuid, 'bin.meshes.idx', 'bin.meshes.idx'),
                (parent_outfolder_uuid, 'bin.energy_mesh', 'bin.energy_mesh'),
                (parent_outfolder_uuid, 'bin.atoms', 'bin.atoms'), (parent_outfolder_uuid, 'bin.dims', 'bin.dims'),
                (parent_outfolder_uuid, self._DEFAULT_OUTPUT_PREP_FILE, self._DEFAULT_OUTPUT_PREP_FILE),
                (parent_outfolder_uuid, self._DEFAULT_OUTPUT_FILE, self._DEFAULT_OUTPUT_FILE)
            ]

        # Check inputdict
        parameters = self._check_input_dict(parameters, num_mpi_procs, convert)

        #Check if parent is a KKRnano calculation
        write_efermi = False
        if not use_strucwithpot:
            self._check_valid_parent(parent_outfolder)
            if parent_calc_calc_node.process_label == 'KKRnanoCalculation':
                fermi = parent_calc_calc_node.outputs.output_parameters.get_dict()['fermi_energy_in_ryd'][-1]
                write_efermi = True

        # Check if non-colinear calculation mode is activated
        noco = False
        if 'KORBIT' in parameters:
            if parameters['KORBIT']['value'] == 1:
                noco = True
        if 'soc' in parameters:
            if parameters['soc']['value'] == True:
                if noco == False:
                    parameters['KORBIT']['value'] = 1
                    self.logger.warn(
                        'KORBIT was set to 1 -> SOC is implemented for the NOCO-Chebyshev solver, only! This is however not changed in the input node and might lead to inconsistencies if this feature has been added in KKRnano!'
                    )
                    noco = True

        if noco:
            with tempfolder.open(self._DEFAULT_NOCO_INPUT_FILE, u'w') as nonco_angles_handle:
                self._write_nonco_angles(nonco_angles_handle, nonco_angles, structure)  # pylint: disable=possibly-used-before-assignment

        # Prepare rbasis.xyz and input.conf from Structure and input parameter data unless convert mode
        if not convert:
            with tempfolder.open(self._DEFAULT_INPUT_FILE, u'w') as input_file_handle:
                self._write_input_file(input_file_handle, parameters, structure, passed_lattice_const)
            with tempfolder.open(self._RBASIS, u'w') as rbasis_handle:
                self._write_rbasis(rbasis_handle, structure, passed_lattice_const)
        if write_efermi:
            with tempfolder.open(self._DEFAULT_EFERMI_FILE, u'w') as efermi_file_handle:
                self._write_efermi_file(efermi_file_handle, fermi)  # pylint: disable=used-before-assignment

        # Prepare potential and shapefun file from strucwithpot, if necessary
        if use_strucwithpot:
            with tempfolder.open(self._POTENTIAL, u'w') as potential_file_handle:
                self._write_potential_file(potential_file_handle, strucwithpot)
            with tempfolder.open(self._SHAPEFUN, u'w') as shapefun_file_handle:
                self._write_shapefun_file(shapefun_file_handle, strucwithpot)

        # Prepare CalcInfo to be returned to aiida
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid

        if not use_strucwithpot:
            calcinfo.local_copy_list = self._get_local_copy_list(parent_calc, local_copy_list_for_continued)
        if use_strucwithpot:
            calcinfo.local_copy_list = []
        calcinfo.remote_copy_list = []

        #retrieve list
        #TODO add NOCO
        calcinfo.retrieve_list = [
            self._DEFAULT_OUTPUT_PREP_FILE, self._DEFAULT_OUTPUT_FILE, self._DEFAULT_NOCO_OUTPUT_FILE, 'bin.dims',
            self._CONVERT_OUTPUT_FILE, 'vpot*', 'DOS*', 'nonco_angle_out.dat'
        ]  # bin.dims is added here, as this should be retrieved for generating vpot-files (negligable in size anyway)

        for j in range(
            len(local_copy_list_for_continued)
        ):  #add the binary files that are necessary to restart a calculation to the retrieved list
            calcinfo.retrieve_list.append(local_copy_list_for_continued[j][1])

        codeinfo = CodeInfo()
        if convert:
            codeinfo.cmdline_params = ['--convert']
            codeinfo.stdout_name = self._CONVERT_OUTPUT_FILE
        else:
            codeinfo.cmdline_params = []
            codeinfo.stdout_name = self._DEFAULT_OUTPUT_FILE
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]

        return calcinfo

    def _list2string(self, list0):
        string = ''
        for item in list0:
            string += str(item) + ' '
        return string

    def _array2string(self, array):
        if array.shape == ():
            return str(array)
        else:
            return self._list2string(array)

    def _getParametersEntry(self, key, value):
        """writing out various entry types from a dict into the format suitable for KKRnano"""

        if type(value) == str or type(value) == int:
            content = f'{key} = {value}'
        elif type(value) == float or type(value) == np.float_:
            content = f"{key} = {str.replace(str(value), 'e', 'D')}"
        elif type(value) == np.ndarray:
            content = f'{key} = {self._array2string(value)}'
        elif type(value) == list:
            content = f'{key} = {self._list2string(value)}'
        elif type(value) is bool:
            value2write = 'f'
            if value:
                value2write = 't'
            self.logger.info('TEST123')
            content = f'{key} = {value2write}'
        else:
            print(f'WARNING: Unknown datatype of entry "{value}".                   Assume array and proceed')
            try:
                value = np.array(value)
                content = f'{key} = {self._array2string(value)}'
                print(f'{key} = {self._array2string(value)}')
                return content
            except:
                self.logger.error('ERROR: Datatype cannot be used as array!')
        content = content + '\n'
        return content

    def _get_lattice_constant(self, structure, passed_lattice_const):
        """determine whether a passed lattice constant should be used"""
        # default value, see above. (Probably) No one would use this large a lattice constant to replace the other
        if passed_lattice_const == -10000.0:
            lattice_param_angs = max(structure.lattice.abc)
        elif passed_lattice_const > 0.0:
            lattice_param_angs = passed_lattice_const
            self.logger.info('Using passed lattice constant. The vectors are scaled accordingly!')
        else:
            raise InputValidationError('The passed lattice constant must not be negative!')
        return lattice_param_angs

    def _write_input_file(self, input_file_handle, parameters, structure, passed_lattice_const):
        """write the input.conf file for KKRnano"""

        #lattice_param_angs=max(structure.lattice.abc)

        lattice_param_angs = self._get_lattice_constant(structure, passed_lattice_const)

        write_list = []

        #header of input.conf
        write_list.append('# LATTICE \n \n# lattice parameter in units of the Bohr radius \n')

        write_list.append(f'alat = {lattice_param_angs * get_Ang2aBohr()}')

        write_list.append('\n# scale basis coordinates by these factors \nbasisscale = 1.0  1.0  1.0 \n \n')
        latt_dict = structure.lattice.as_dict()

        write_list.append('#BRAVAIS\n')
        rbrav = np.array(latt_dict['matrix']) / lattice_param_angs

        directions = ['a', 'b', 'c']
        for i in range(3):
            write_list.append(f'bravais_{directions[i]}=  {rbrav[i][0]:.15f}  {rbrav[i][1]:.15f}  {rbrav[i][2]:.15f}\n')
        write_list.append('\ncartesian = t\n')
        #writing other parameters from parameter dict
        #params=parameters.get_dict()
        params = parameters

        for key in params:
            write_list.append(self._getParametersEntry(key, params[key]['value']))

        write_list.append('# CHANGED SCRIPT')

        input_file_handle.writelines(write_list)

    def _write_nonco_angles(self, nonco_angles_handle, nonco_angles, structure):
        """
        write nonco_angles.dat file for KKRnano
        created from dictionary with structure dict={'atom':{1:{'theta':0.0,'phi':00.0, 'fix_angle_mode':1},...}
        The angles 'theta' (polar angle going from z- to x-direction) and 'phi' (azimuthal angle) are given in deg.
        'fix_angle_mode' takes values 0,1,2,3. 0 is for relaxation of the spin-direciton, 1 is for fixing it; 2 and 3
        are for constraining fields calculations.
        """
        n_atoms = len(structure.atomic_numbers)

        #TODO: Adapt to KKRhost style
        #Create dictionary if none was given in the input
        if nonco_angles == {}:
            nonco_angles = {'atom': {}}
            for i in range(n_atoms):
                nonco_angles['atom'][i + 1] = {'theta': 0.0, 'phi': 0.0, 'fix_angle_mode': 1}
        if n_atoms != len(nonco_angles['atom'].keys()):
            raise InputValidationError(
                'The number of atoms in the structure must\
            match the number of atoms specified for the non-colinear angles.'
            )

        write_list = []
        for key in nonco_angles['atom']:
            write_list.append('{} {} {}\n'.format(nonco_angles['atom'][key]['theta'],\
                                                  nonco_angles['atom'][key]['phi'],\
                                                  nonco_angles['atom'][key]['fix_angle_mode']))

        nonco_angles_handle.writelines(write_list)

    def _get_rbasis_atom_symbol(self, atomic_number):
        """returns either an element symbol or a vacuum symbol"""
        #PSE element symbols
        symbols = (
            '__', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',
            'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
            'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
            'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf',
            'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
            'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs',
            'Mt', 'Ds', 'Rg', 'Cn', 'Uut', 'Fl', 'Uup', 'Lv', 'Uus', 'Uuo'
        )
        if abs(atomic_number) > len(
            symbols
        ) + 30:  #pymatgen stores X sites with very high random `atomic numbers` that can also be negative
            return symbols[0]
        else:
            return symbols[atomic_number]

    def _write_rbasis(self, rbasis_handle, structure, passed_lattice_const):
        """write the rbasis.xyz file for KKRnano"""

        try:
            space_group_info = str(structure.get_space_group_info())
        except:
            space_group_info = 'No space group could be identifed'

        write_list=[str(len(structure.atomic_numbers)),'\n',\
                    str(structure.composition), ', ', space_group_info,'\n']

        lattice_param_angs = self._get_lattice_constant(structure, passed_lattice_const)

        for i in range(np.shape(structure.cart_coords)[0]):
            write_list.append('{}   {:.15f}   {:.15f}   {:.15f}\n'.format(self._get_rbasis_atom_symbol(structure.atomic_numbers[i]), \
               structure.cart_coords[i,0]/lattice_param_angs, \
               structure.cart_coords[i,1]/lattice_param_angs,\
               structure.cart_coords[i,2]/lattice_param_angs))

        rbasis_handle.writelines(write_list)

    def _write_efermi_file(self, efermi_handle, efermi):
        """ write file EFERMI necessary to restart a calculation. """
        efermi_handle.writelines([str(efermi)])

    def _write_potential_file(self, potential_handle, strucwithpot):
        """ write file potential from strucwithpot object's potential files. """
        for vpot in strucwithpot.potentials:
            potential_handle.writelines(vpot.get_content())  #.split('\n'))

    def _write_shapefun_file(self, shapefun_handle, strucwithpot):
        """ write file potential from strucwithpot object's potential files. """

        shapefun_handle.writelines(strucwithpot.make_shapefun())  #.split('\n'))

    def _get_local_copy_list(self, parent_calc_remote, local_copy_list_for_continued):
        """find files to copy from the parent_folder's retrieved to the input of a KKRnano calculation"""
        #Note: the check if a non-empty copylist is needed, i. e. if started
        # from a parent calculation, is done in prepare_for_submission
        parent_calc = parent_calc_remote.get_incoming(node_class=CalcJobNode).first().node
        retrieved = parent_calc.outputs.retrieved

        local_copy_list = []
        if not self._is_KkrnanoCalc(parent_calc):
            # copy input potential from voronoi output
            local_copy_list += [(retrieved.uuid, parent_calc.process_class._OUT_POTENTIAL_voronoi, self._POTENTIAL)]

            # copy shapefun from voronoi output
            local_copy_list += [(retrieved.uuid, parent_calc.process_class._SHAPEFUN, self._SHAPEFUN)]

        elif parent_calc.process_label == 'KKRnanoCalculation':
            local_copy_list = local_copy_list_for_continued
        return local_copy_list

    def _check_valid_parent(self, parent_calc_folder):
        """
        Check that calc is a valid parent for a KKRnano.
        It can be a VoronoiCalculation, KKRnanoCalculation
        """

        # extract parent calculation
        parent_calcs = parent_calc_folder.get_incoming(node_class=CalcJobNode)
        n_parents = len(parent_calcs.all_link_labels())
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
        else:
            parent_calc = parent_calcs.first().node
            overwrite_pot = True


        if (not parent_calc.process_label=='KKRnanoCalculation')\
    and (not parent_calc.process_label=='VoronoiCalculation'):
            raise ValueError('Parent Calculation has to be another KKRnano or a Voronoi calculation.')

    def _check_input_dict(self, inputdict, num_mpi_procs, convert):
        """ checks if all essential keys are contained in the inputdict and if it has the right format """
        #Check if all essential keys are present
        all_present = True
        missing_keys = []
        for key in self._DEFAULT_KKRNANO_PARA:
            if not key in inputdict:
                all_present = False
                missing_keys.append(key)
        if not all_present:
            raise InputValidationError(
                'Not all essential keys were given in the input dictionary. \
            At least the following keys are missing: \
            {}'.format(missing_keys)
            )
        #Check for format
        keys_in_wrong_format = []
        correct_format = True
        for key in inputdict:
            try:
                if not 'value' in inputdict[key]:
                    correct_format = False
                    keys_in_wrong_format.append(key)
            except:
                correct_format = False
                keys_in_wrong_format.append(key)
        if not correct_format:
            self.logger.warn(
                'Some keys were provided in an incorrect format. Dict entries must have entry `value`\
(see e. g. KKRnano-Plugin/aiida-kkr/aiida_kkr/calculations/kkrnano.py). At least the following keys are in an incorrect format: \
{} \nMoving on by assuming just the key `value` was missing and adapt accordingly'.format(keys_in_wrong_format)
            )
            #try to correct incorrect entries, most likely they do not have the 'value' feature
            try:
                for key in keys_in_wrong_format:
                    likely_value = inputdict[key]
                    inputdict[key] = {'value': likely_value}
            except:
                raise InputValidationError(
                    'Some keys were provided in an incorrect format. Dict entries must have entry `value`\
(see e. g. KKRnano-Plugin/aiida-kkr/aiida_kkr/calculations/kkrnano.py). At least the following keys are in an incorrect format: \
{}.'.format(keys_in_wrong_format)
                )

        #Check if the values are provided in the correct data type
        all_correct_dtype = True
        keys_with_incorrect_dtypes = []
        values_with_incorrect_dtypes = []
        correct_dtypes_for_incorrect_keys = []
        for key in self._DEFAULT_KKRNANO_PARA:
            default_value = self._DEFAULT_KKRNANO_PARA[key]['value']
            dtype_required = type(default_value)

            entered_value = inputdict[key]['value']
            dtype_entered = type(entered_value)

            if dtype_required == list:
                try:
                    for item_index in range(len(default_value)):
                        default_entry = default_value[item_index]
                        entered_entry = entered_value[item_index]
                        if not type(default_entry) == type(entered_entry):
                            all_correct_dtype = False
                            keys_with_incorrect_dtypes.append(key)
                            values_with_incorrect_dtypes.append(str(entered_value))
                            correct_dtypes_for_incorrect_keys.append(
                                f'array/list of {str(type(default_entry))} (wrong entry type)'
                            )
                except IndexError:
                    all_correct_dtype = False
                    keys_with_incorrect_dtypes.append(key)
                    values_with_incorrect_dtypes.append(str(entered_value))
                    correct_dtypes_for_incorrect_keys.append(
                        'array/list of {} (wrong length, \
length should be {} )'.format(str(type(default_entry)), str(len(default_value)))
                    )

            elif not dtype_required == dtype_entered:
                all_correct_dtype = False
                keys_with_incorrect_dtypes.append(key)
                values_with_incorrect_dtypes.append(str(entered_value))
                correct_dtypes_for_incorrect_keys.append(f'{str(type(default_entry))}')

            if not all_correct_dtype:
                #Generate helpful error message
                dtype_error_message = 'Some keys were provided with an incorrect datatype. These values are \n'
                for j in range(len(keys_with_incorrect_dtypes)):
                    dtype_error_message+='key: {}, entered value: {}, needed datatype: {}\n'.format(keys_with_incorrect_dtypes[j],\
                                                values_with_incorrect_dtypes[j],correct_dtypes_for_incorrect_keys[j])
                raise InputValidationError(dtype_error_message)

        #Check if the number of mpi processes, specified with the builder matches the settings in the inputdict
        mpikeys = ['NTHRDS', 'EMPID', 'SMPID', 'num_atom_procs']
        needed_num_mpi_procs = 1
        for key in mpikeys:
            if key in inputdict:
                needed_num_mpi_procs *= inputdict[key]['value']
        if num_mpi_procs % needed_num_mpi_procs != 0 and not convert:
            raise (
                InputValidationError(
                    "Number of MPI processes does not match! In builder.metadata.options.resources['num_mpi_procs']= {}, however according to the builder.parameters node, the needed number of MPI process is {}"
                    .format(num_mpi_procs, needed_num_mpi_procs)
                )
            )
        elif convert and num_mpi_procs != 1:
            #TODO check if convert step works with more cores
            raise (
                InputValidationError(
                    "Number of MPI processes does not match! In builder.metadata.options.resources['num_mpi_procs']= {}, however for the convert step the needed number of MPI processes is 1."
                    .format(num_mpi_procs)
                )
            )

        return inputdict

    def _is_KkrnanoCalc(self, calc):
        """
        check if calc contains the file out_potential
        """
        is_KKR = False
        if calc.process_type == 'aiida.calculations:kkr.kkrnano':
            retrieved_node = calc.get_retrieved_node()
            if True:  #'out_potential' in retrieved_node.list_object_names():
                is_KKR = True

        return is_KKR

    @classmethod
    def _get_struc(self, parent_calc):
        """
        Get strucwithpot from a parent_folder (result of a calculation, typically a remote folder)
        """
        try:
            return parent_calc.inputs.strucwithpot.structure
        except:
            return parent_calc.inputs.structure

    @classmethod
    def _has_struc(self, parent_folder):
        """
        Check if parent_folder has strucwithpot information in its input
        """
        success = True
        link_labels = parent_folder.get_incoming().all_link_labels()
        if 'strucwithpot' not in link_labels and 'structure' not in link_labels:
            success = False
        return success

    @classmethod
    def find_parent_struc_from_voro_or_stwpd(self, parent_folder):
        """
        Find the StructureData node recuresively in chain of parent calculations (structure node is input to voronoi calculation or in input StrucWithPotData node of KKRnano calculation)

        returns structure, parent_folder
        """
        iiter = 0
        Nmaxiter = 1000
        parent_folder_tmp = get_remote(parent_folder)
        while not self._has_struc(parent_folder_tmp) and iiter < Nmaxiter:
            parent_folder_tmp = get_remote(get_parent(parent_folder_tmp))
            iiter += 1
            if iiter % 200 == 0:
                print(
                    'Warning: find_parent_structure takes quite long (already searched {} ancestors). Stop after {}'.
                    format(iiter, Nmaxiter)
                )
        if self._has_struc(parent_folder_tmp):
            struc = self._get_struc(parent_folder_tmp)
            return struc, parent_folder_tmp
        else:
            raise ValueError(f'structure not found {parent_folder_tmp}')



# -*- coding: utf-8 -*-
"""
Helper calcfunction that allws to modify a potential file
"""

from aiida.engine import calcfunction
from aiida.common.exceptions import InputValidationError
from aiida_kkr.tools.context import open_context_to_stack
from contextlib import ExitStack


@calcfunction
def neworder_potential_wf(settings_node, parent_calc_folder, **kwargs):
    """
    Workfunction to create database structure for aiida_kkr.tools.modify_potential.neworder_potential function
    A temporary file is written in a Sandbox folder on the computer specified via
    the input computer node before the output potential is stored as SinglefileData
    in the Database.

    :param settings_node: settings for the neworder_potential function (Dict)
    :param parent_calc_folder: parent calculation remote folder node where the input
        potential is retreived from (RemoteData)
    :param parent_calc_folder2: *optional*, parent calculation remote folder node where
        the second input potential is retreived from in case 'pot2' and 'replace_newpos'
        are also set in settings_node (RemoteData)
    :param debug: *optional*, contol wether or not debug information is written out (aiida.orm.Bool)

    :returns: output_potential node (SinglefileData)

    .. note::

        The settings_node dictionary needs to be of the following form::

            settings_dict = {'neworder': [list of intended order in output potential]}

        Optional entries are::

            'out_pot': '<filename_output_potential>'  name of the output potential file, defaults to 'potential_neworder' if not specified
            'pot1': '<filename_input_potential>'      if not given we will try to find it from the type of the parent remote folder
            'pot2': '<filename_second_input_file>'
            'replace_newpos': [[position in neworder list which is replace with potential from pot2, position in pot2 that is chosen for replacement]]
            'switch_spins': [indices of atom for which spins are exchanged] (indices refer to position in neworder input list)
            'label': 'label_for_output_node'
            'description': 'longer_description_for_output_node'
    """
    import os
    from aiida_kkr.tools.tools_kkrimp import modify_potential
    from aiida.common.folders import SandboxFolder
    from aiida.common.exceptions import UniquenessError
    from aiida.orm import CalcJobNode, Dict, RemoteData, SinglefileData

    if 'debug' in list(kwargs.keys()):
        debug = kwargs.get('debug').value
    else:
        debug = False

    if 'parent_calc_folder2' in list(kwargs.keys()):
        parent_calc_folder2 = kwargs.get('parent_calc_folder2', None)
    else:
        parent_calc_folder2 = None

    # check input consistency
    if not isinstance(settings_node, Dict):
        raise InputValidationError('settings_node needs to be a valid aiida Dict node')
    if not isinstance(parent_calc_folder, RemoteData):
        raise InputValidationError('parent_calc_folder needs to be a valid aiida RemoteData node')
    if parent_calc_folder2 is not None and not isinstance(parent_calc_folder2, RemoteData):
        raise InputValidationError('parent_calc_folder2 needs to be a valid aiida RemoteData node')

    settings_dict = settings_node.get_dict()
    pot1 = settings_dict.get('pot1', None)
    if pot1 is None:
        # try to extract the potential name from the type of the parent_calc_folder
        try:
            pot1 = extract_potname_from_remote(parent_calc_folder)
        except ValueError:
            raise InputValidationError(
                'settings_node_dict needs to have key "pot1" containing the filename of the input potential'
            )
    out_pot = settings_dict.get('out_pot', 'potential_neworder')
    neworder = settings_dict.get('neworder', None)
    if neworder is None:
        raise InputValidationError(
            'settings_node_dict needs to have key "neworder" containing the list of new positions'
        )
    pot2 = settings_dict.get('pot2', None)
    if pot2 is None and parent_calc_folder2 is not None:
        # try to extract the potential name from the type of the parent_calc_folder
        try:
            pot2 = extract_potname_from_remote(parent_calc_folder2)
        except ValueError:
            raise InputValidationError(
                'settings_node_dict needs to have key "pot2" containing the filename of the input potential'
            )
    replace_newpos = settings_dict.get('replace_newpos', None)
    switch_spins = settings_dict.get('switch_spins', [])

    # Create Sandbox folder for generation of output potential file
    # and construct output potential
    with SandboxFolder() as tempfolder:
        # Get abolute paths of input files from parent calc and filename
        parent_calcs = parent_calc_folder.get_incoming(node_class=CalcJobNode).all()
        n_parents = len(parent_calcs)
        if n_parents != 1:
            raise UniquenessError(
                'Input RemoteData is child of {} '
                'calculation{}, while it should have a single parent'
                ''.format(n_parents, '' if n_parents == 0 else 's')
            )
        parent_calc = parent_calcs[0].node

        # extract nspin from parent calc's input parameter node
        nspin = parent_calc.inputs.parameters.get_dict().get('NSPIN')
        neworder_spin = []
        ii = 0
        for iatom in neworder:
            spins = range(nspin)
            # change spin order if needed
            if ii in switch_spins:
                spins = spins[::-1]
            for ispin in spins:
                neworder_spin.append(iatom * nspin + ispin)
            ii += 1
        neworder = neworder_spin

        # Copy optional files?
        if pot2 is not None and parent_calc_folder2 is not None:
            parent_calcs = parent_calc_folder2.get_incoming(node_class=CalcJobNode).all()
            n_parents = len(parent_calcs)
            if n_parents != 1:
                raise UniquenessError(
                    'Input RemoteData of parent_calc_folder2 is child of {} '
                    'calculation{}, while it should have a single parent'
                    ''.format(n_parents, '' if n_parents == 0 else 's')
                )
            else:
                parent_calc2 = parent_calcs[0].node
            if pot2 not in parent_calc2.outputs.retrieved.list_object_names():
                raise InputValidationError(
                    'neworder_potential_wf: pot2 does not exist', pot2,
                    parent_calc.outputs.retrieved.list_object_names()
                )

        # open files in context (if name is None the fhandle will just be None and then ignored)
        with ExitStack() as stack:
            out_pot_fhandle = open_context_to_stack(stack, tempfolder, out_pot, u'w')
            pot1_fhandle = open_context_to_stack(stack, parent_calc.outputs.retrieved, pot1)
            pot2_fhandle = None
            if pot2 is not None:
                pot2_fhandle = open_context_to_stack(stack, parent_calc2.outputs.retrieved, pot2)
            # run neworder_potential function
            modify_potential().neworder_potential(
                pot1_fhandle,
                out_pot_fhandle,
                neworder,
                potfile_2=pot2_fhandle,
                replace_from_pot2=replace_newpos,
                debug=debug
            )

        # store output potential to SinglefileData
        with tempfolder.open(out_pot, u'rb') as fhandle:
            output_potential_sfd_node = SinglefileData(file=fhandle)

        lbl = settings_dict.get('label', None)
        if lbl is not None:
            output_potential_sfd_node.label = lbl
        desc = settings_dict.get('description', None)
        if desc is not None:
            output_potential_sfd_node.description = desc

        # TODO create shapefun sfd node accordingly
        """
        out_shape_path =

        output_shapefun_sfd_node = SinglefileData(file=out_shape_path)

        lbl2 = settings_dict.get('label_shape', None)
        if lbl2 is None and lbl is not None:
            lbl2 = lbl
        if lbl2 is not None:
            output_shapefun_sfd_node.label = lbl2
        desc2 = settings_dict.get('description_shape', None)
        if desc2 is None and desc is not None:
            desc2 = desc
        if desc2 is not None:
            output_shapefun_sfd_node.description = desc2

        return output_potential_sfd_node, output_shapefun_sfd_node
        """
        return output_potential_sfd_node


def extract_potname_from_remote(parent_calc_folder):
    """
    extract the bname of the output potential from a RemoteData folder
    """
    from aiida_kkr.calculations import KkrCalculation, VoronoiCalculation
    from aiida.orm import CalcJobNode

    pot_name = None
    # extract list of parents (can only extract the parent calculation
    # if there is only a single incoming link to follow)
    parents = parent_calc_folder.get_incoming(node_class=CalcJobNode)
    if len(list(parents)) == 1:
        parent = parents.first().node
        # now extract the pot_name dependeing on the parent calculation's type
        if parent.process_class == KkrCalculation:
            pot_name = KkrCalculation._OUT_POTENTIAL
        elif parent.process_class == VoronoiCalculation:
            pot_name = VoronoiCalculation._OUT_POTENTIAL_voronoi

    # return the potential name or raise an error if nothing was found
    if pot_name is not None:
        return pot_name
    else:
        raise ValueError('Could not extract a potential name')



# -*- coding: utf-8 -*-
"""
Here workfunctions and normal functions using aiida-stuff (typically used
within workfunctions) are collected.
"""

from aiida.common.exceptions import InputValidationError
from aiida.engine import calcfunction
from aiida.orm import Dict
from masci_tools.io.kkr_params import kkrparams
from builtins import str

# keys that are used by aiida-kkr some something else than KKR parameters
_ignored_keys = ['ef_set', 'use_input_alat', '<NEWVERSION_BDG>', '<DECOUPLE_SPINS_CHEBY>']
_ignored_keys += [i.upper() for i in _ignored_keys]


@calcfunction
def update_params_wf(parameternode, updatenode, **link_inputs):
    """
    Work function to update a KKR input parameter node.
    Stores new node in database and creates a link from old parameter node to new node
    Returns updated parameter node using update_params function

    :note: Input nodes need to be valid aiida Dict objects.

    :param parameternode: Input aiida Dict node cotaining KKR specific parameters
    :param updatenode: Input aiida Dict node containing a dictionary with the parameters that are supposed to be changed.

    :note: If 'nodename' is contained in dict of updatenode the string corresponding to this key will be used as nodename for the new node. Otherwise a default name is used
    :note: Similar for 'nodedesc' which gives new node a description

    :example: updated_params = Dict(dict={'nodename': 'my_changed_name', 'nodedesc': 'My description text', 'EMIN': -1, 'RMAX': 10.})
              new_params_node = update_params_wf(input_node, updated_params)
    """
    updatenode_dict = updatenode.get_dict()
    if 'nodename' in list(updatenode_dict.keys()):
        # take nodename out of dict (should only contain valid KKR parameter)
        nodename = updatenode_dict.pop('nodename')
    else:
        nodename = None
    if 'nodedesc' in list(updatenode_dict.keys()):
        # take nodename out of dict (should only contain valid KKR parameter later on)
        nodedesc = updatenode_dict.pop('nodedesc')
    else:
        nodedesc = None

    # do nothing if updatenode is empty
    if len(list(updatenode_dict.keys())) == 0:
        print('Input node is empty, do nothing!')
        raise InputValidationError('Nothing to store in input')
    #
    new_parameternode = update_params(parameternode, nodename=nodename, nodedesc=nodedesc, **updatenode_dict)
    return new_parameternode


def update_params(node, nodename=None, nodedesc=None, strict=False, **kwargs):
    """
    Update parameter node given with the values given as kwargs.
    Returns new node.

    :param node: Input parameter node (needs to be valid KKR input parameter node).
    :param **kwargs: Input keys with values as in kkrparams.
    :param linkname: Input linkname string. Give link from old to new node a name .
                     If no linkname is given linkname defaults to 'updated parameters'

    :return: parameter node

    :example usage: OutputNode = KkrCalculation.update_params(InputNode, EMIN=-1, NSTEPS=30)

    :note: Keys are set as in kkrparams class. Check documentation of kkrparams for further information.
    :note: If kwargs contain the key `add_direct`, then no kkrparams instance is used and no checks are performed but the dictionary is filled directly!
    :note: By default nodename is 'updated KKR parameters' and description contains list of changed
    """
    # check if node is a valid KKR parameters node
    if not isinstance(node, Dict):
        print('Input node is not a valid Dict node')
        raise InputValidationError('update_params needs valid parameter node as input')

    # check if add_direct is in kwargs (shortcuts checks of kkrparams by not using the kkrparams class to set the dict)
    add_direct = False
    if 'add_direct' in list(kwargs.keys()):
        add_direct = kwargs.pop('add_direct')

    # initialize temporary kkrparams instance containing all possible KKR parameters
    if not add_direct:
        params = kkrparams()
    else:
        params = {}

    # extract input dict from node
    inp_params = node.get_dict()

    # check if input dict contains only values for KKR parameters
    if not add_direct:
        remove_keys = []
        for key in inp_params:
            if key not in list(params.values.keys()) and key not in _ignored_keys:
                print(f'WARNING: Input node contains invalid key "{key}"')
                if strict:
                    raise InputValidationError(f'invalid key "{key}" in input parameter node')
                else:
                    # print a warning and remove the key
                    print(f'ignore this key/value pair: {key}: {inp_params.get(key)}')
                    remove_keys.append(key)
        for key in remove_keys:
            inp_params.pop(key)

    # copy values from input node
    for key in inp_params:
        value = inp_params[key]
        if not add_direct:
            params.set_value(key, value, silent=True)
        else:
            params[key] = value

    # to keep track of changed values:
    changed_params = {}

    # check if values are given as **kwargs (otherwise return input node)
    if len(kwargs) == 0:
        print('No additional input keys given, return input node')
        return node.clone()
    for key in kwargs:
        # check if value of 'key' should be set (either because it differs from old para node or because it was not set at all)
        update_value = False
        if key in list(inp_params.keys()):
            if kwargs[key] != inp_params[key]:
                update_value = True
        else:
            update_value = True
        if update_value:
            if not add_direct:
                params.set_value(key, kwargs[key], silent=True)
            else:
                params[key] = kwargs[key]
            changed_params[key] = kwargs[key]

    if len(list(changed_params.keys())) == 0:
        print('No keys have been changed, return input node')
        return node.clone()

    # set linkname with input or default value
    if nodename is None or not isinstance(nodename, str):
        nodename = 'updated KKR parameters'
    if nodedesc is None or not isinstance(nodedesc, str):
        nodedesc = f'changed parameters: {changed_params}'

    # create new node
    if not add_direct:
        ParaNode = Dict(params.values)
    else:
        ParaNode = Dict(params)
    ParaNode.label = nodename
    ParaNode.description = nodedesc

    return ParaNode


# TODO implment VCA functionality
# maybe one starts from a calculation closest to the VCA case and slowly
# increase ZATOM which violates the _do_never_modify rule in KKR calculation
# this should then create a new structure and modify the old potential accordingly
# general rule: Nover destroy the data provenance!!!


@calcfunction
def prepare_VCA_structure_wf():
    pass


def prepare_VCA_structure():
    pass


# TODO implement 2D input helper
# a helper workfunction would be nice to create the vacuum region etc. for 2D calculation
@calcfunction
def prepare_2Dcalc_wf():
    pass


def prepare_2Dcalc():
    pass


def test_and_get_codenode(codenode, expected_code_type, use_exceptions=False):
    """
    Pass a code node and an expected code (plugin) type. Check that the
    code exists, is unique, and return the Code object.

    :param codenode: the name of the code to load (in the form label@machine)
    :param expected_code_type: a string with the plugin that is expected to
      be loaded. In case no plugins exist with the given name, show all existing
      plugins of that type
    :param use_exceptions: if True, raise a ValueError exception instead of
      calling sys.exit(1)
    :return: a Code object

    :example usage: from kkr_scf workflow::

        if 'voronoi' in inputs:
            try:
                test_and_get_codenode(inputs.voronoi, 'kkr.voro', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for voronoi  does not "
                         "use the plugin kkr.voro")
                self.control_end_wc(error)
    """
    import sys
    from aiida.common.exceptions import NotExistent
    from aiida.orm import Code

    try:
        if codenode is None:
            raise ValueError
        code = codenode
        if code.get_input_plugin_name() != expected_code_type:
            raise ValueError
    except (NotExistent, ValueError):
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(Code, filters={'attributes.input_plugin': {'==': expected_code_type}}, project='*')

        valid_code_labels = [f'{c.label}@{c.computer.label}' for [c] in qb.all()]

        if valid_code_labels:
            msg = (
                'Pass as further parameter a valid code label.\n'
                'Valid labels with a {} executable are:\n'.format(expected_code_type)
            )
            msg += '\n'.join(f'* {label}' for label in valid_code_labels)

            if use_exceptions:
                raise ValueError(msg)
            else:
                print(msg, file=sys.stderr)
                sys.exit(1)
        else:
            msg = (
                'Code not valid, and no valid codes for {}.\n'
                'Configure at least one first using\n'
                '    verdi code setup'.format(expected_code_type)
            )
            if use_exceptions:
                raise ValueError(msg)
            print(msg, file=sys.stderr)
            sys.exit(1)

    return code


def get_inputs_kkr(code, remote, options, label='', description='', parameters=None, serial=False, imp_info=None):
    """
    Get the input for a voronoi calc.
    Wrapper for KkrProcess setting structure, code, options, label, description etc.
    :param code: a valid KKRcode installation (e.g. input from Code.get_from_string('codename@computername'))
    :param remote: remote directory of parent calculation (Voronoi or previous KKR calculation)

    """
    from aiida_kkr.calculations.kkr import KkrCalculation

    # then reuse common inputs setter
    builder = get_inputs_common(
        KkrCalculation, code, remote, None, options, label, description, parameters, serial, imp_info
    )

    return builder


def get_inputs_kkrimporter(code, remote, options, label='', description='', parameters=None, serial=False):
    """
    Get the input for a voronoi calc.
    Wrapper for KkrProcess setting structure, code, options, label, description etc.
    """
    from aiida_kkr.calculations.kkr import KkrCalculation
    KkrProcess = KkrCalculation.process()

    # then reuse common inputs setter
    inputs = get_inputs_common(KkrProcess, code, remote, None, options, label, description, parameters, serial)

    return inputs


def get_inputs_voronoi(code, structure, options, label='', description='', params=None, serial=True, parent_KKR=None):
    """
    Get the input for a voronoi calc.
    Wrapper for VoronoiProcess setting structure, code, options, label, description etc.
    """
    # get process for VoronoiCalculation
    from aiida_kkr.calculations.voro import VoronoiCalculation

    # then reuse common inputs setter all options
    if structure is not None:
        # for 'normal' case starting from structure
        builder = get_inputs_common(
            VoronoiCalculation, code, None, structure, options, label, description, params, serial
        )
    else:
        # for parent_KKR feature used to increase lmax which cannot have 'structure' in inputs
        builder = get_inputs_common(
            VoronoiCalculation, code, None, None, options, label, description, params, serial, parent_KKR=parent_KKR
        )

    return builder


def get_inputs_kkrimp(
    code,
    options,
    label='',
    description='',
    parameters=None,
    serial=False,
    imp_info=None,
    host_GF=None,
    imp_pot=None,
    kkrimp_remote=None,
    host_GF_Efshift=None
):
    """
    Get the input for a kkrimp calc.
    Wrapper for KkrimpProcess setting structure, code, options, label, description etc.
    :param code: a valid KKRimpcode installation (e.g. input from Code.get_from_string('codename@computername'))
    TBD
    """

    from aiida_kkr.calculations.kkrimp import KkrimpCalculation

    # then reuse common inputs setter
    builder = get_inputs_common(
        KkrimpCalculation, code, None, None, options, label, description, parameters, serial, imp_info, host_GF,
        imp_pot, kkrimp_remote, host_GF_Efshift
    )

    return builder


def truncate_string(string, max_length):
    """
    Truncate the length of a string to max_length-3 entries.
    The last three characters '...' to indicate the truncation.
    """
    if len(string) > max_length:
        string = string[:max_length - 3] + '...'
    return string


def get_inputs_common(
    calculation,
    code,
    remote,
    structure,
    options,
    label,
    description,
    params,
    serial,
    imp_info=None,
    host_GF=None,
    imp_pot=None,
    kkrimp_remote=None,
    host_GF_Efshift=None,
    **kwargs
):
    """
    Base function common in get_inputs_* functions for different codes
    """
    inputs = calculation.get_builder()

    if structure:
        inputs.structure = structure

    if remote:
        inputs.parent_folder = remote

    if code:
        inputs.code = code
        _sched = code.computer.scheduler_type
    else:
        _sched = None
    if params:
        inputs.parameters = params

    if not options:
        options = {}

    if description:
        inputs.metadata.description = description
    else:
        inputs.metadata.description = ''

    if label:
        # Attention: max label length is 255 characters
        inputs.metadata.label = truncate_string(label, 255)
    else:
        inputs.metadata.label = ''

    if serial:
        # check for old aiida name (e.g. "slurm") and new aiida name ("core.slurm") of the scheduler
        if _sched in ['core.slurm', 'slurm', 'core.pbspro', 'pbspro']:
            # overwrite settings for serial run
            options['withmpi'] = False
            options['resources'] = {'num_machines': 1, 'tot_num_mpiprocs': 1}
        if _sched in ['core.sge', 'sge']:
            options['withmpi'] = False
            options['resources'] = {'parallel_env': 'smpslots', 'tot_num_mpiprocs': 1}
    else:
        # otherwise assume MPI parallelism if not given in input options
        if 'withmpi' not in list(options.keys()):
            options['withmpi'] = True

    if options:
        inputs.metadata.options = options
    '''
    options = {
    "max_wallclock_seconds": int,
    "resources": dict,
    "custom_scheduler_commands": unicode,
    "queue_name": basestring,
    "computer": Computer,
    "withmpi": bool,
    "mpirun_extra_params": Any(list, tuple),
    "import_sys_environment": bool,
    "environment_variables": dict,
    "priority": unicode,
    "max_memory_kb": int,
    "prepend_text": unicode,
    "append_text": unicode}
    '''

    # for kkrimp calculations
    if imp_info is not None:
        inputs.impurity_info = imp_info

    if host_GF is not None:
        inputs.host_Greenfunction_folder = host_GF

    if host_GF_Efshift is not None:
        inputs.host_Greenfunction_folder_Efshift = host_GF_Efshift

    if imp_pot is not None:
        inputs.impurity_potential = imp_pot

    if kkrimp_remote is not None:
        inputs.parent_calc_folder = kkrimp_remote

    # add additional inputs
    for link_label, node in kwargs.items():
        inputs[link_label] = node

    return inputs


def get_parent_paranode(remote_data):
    """
    Return the input parameter of the parent calculation giving the remote_data node
    """
    inp_calc = remote_data.get_incoming(link_label_filter='remote_folder').first().node
    inp_para = inp_calc.get_incoming(link_label_filter='parameters').first().node
    return inp_para


def generate_inputcard_from_structure(
    parameters,
    structure,
    input_filename,
    parent_calc=None,
    shapes=None,
    isvoronoi=False,
    use_input_alat=False,
    vca_structure=False
):
    """
    Takes information from parameter and structure data and writes input file 'input_filename'

    :param parameters: input parameters node containing KKR-related input parameter
    :param structure: input structure node containing lattice information
    :param input_filename: input filename, typically called 'inputcard'

    optional arguments
    :param parent_calc: input parent calculation node used to determine if EMIN
                        parameter is automatically overwritten (from voronoi output)
                        or not
    :param shapes: input shapes array (set automatically by
                   aiida_kkr.calculations.Kkrcalculation and shall not be overwritten)
    :param isvoronoi: tell whether or not the parameter set is for a voronoi calculation or kkr calculation (have different lists of mandatory keys)
    :param use_input_alat: True/False, determines whether the input alat value is taken or the new alat is computed from the Bravais vectors


    :note: assumes valid structure and parameters, i.e. for 2D case all necessary
           information has to be given. This is checked with function
           'check_2D_input' called in aiida_kkr.calculations.Kkrcalculation
    """

    from aiida.common.constants import elements as PeriodicTableElements
    from numpy import array
    from masci_tools.io.kkr_params import kkrparams
    from masci_tools.io.common_functions import get_Ang2aBohr, get_alat_from_bravais
    from aiida_kkr.calculations.voro import VoronoiCalculation

    # initialize list of warnings
    warnings = []

    # list of globally used constants
    a_to_bohr = get_Ang2aBohr()

    # Get the connection between coordination number and element symbol
    # maybe do in a different way

    _atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}

    # KKR wants units in bohr
    bravais = array(structure.cell) * a_to_bohr
    alat_input = parameters.get_dict().get('ALATBASIS')
    if use_input_alat and alat_input is not None:
        alat = alat_input
        wmess = 'found alat in input parameters, this will trigger scaling of RMAX, GMAX and RCLUSTZ!'
        print(f'WARNING: {wmess}')
        warnings.append(wmess)
    else:
        alat = get_alat_from_bravais(bravais, is3D=structure.pbc[2])
    bravais = bravais / alat

    sites = structure.sites
    naez = len(sites)
    positions = []
    charges = []
    weights = []  # for CPA
    isitelist = []  # counter sites array for CPA
    isite = 0
    for site in sites:
        pos = site.position
        # TODO maybe convert to rel pos and make sure that type is right for script (array or tuple)
        abspos = array(pos) * a_to_bohr / alat  # also in units of alat
        positions.append(abspos)
        isite += 1
        sitekind = structure.get_kind(site.kind_name)
        for ikind in range(len(sitekind.symbols)):
            site_symbol = sitekind.symbols[ikind]
            if sitekind.is_alloy:
                wght = sitekind.weights[ikind]
            else:
                wght = 1.
            if not sitekind.has_vacancies:
                zatom_tmp = _atomic_numbers[site_symbol]
            else:
                zatom_tmp = 0.0
            if vca_structure and ikind > 0 and not isvoronoi:
                # for VCA case take weighted average (only for KKR code, voronoi code uses zatom of first site for dummy calculation)
                zatom = zatom * wght_last + zatom_tmp * wght
                # also reset weight to 1
                wght = 1.
            else:
                zatom = zatom_tmp
                if vca_structure and isvoronoi:
                    wght = 1.

            wght_last = wght  # for VCA mode

            # make sure that for VCA only averaged position is written (or first for voronoi code)
            if ((
                vca_structure and ((len(sitekind.symbols) == 1) or (not isvoronoi and ikind == 1) or
                                   (isvoronoi and ikind == 0))
            ) or (not vca_structure)):
                charges.append(zatom)
                weights.append(wght)
                isitelist.append(isite)

    weights = array(weights)
    isitelist = array(isitelist)
    charges = array(charges)
    positions = array(positions)

    ######################################
    # Prepare keywords for kkr from input structure

    # get parameter dictionary
    input_dict = parameters.get_dict()

    # remove special keys that are used for special cases but are not part of the KKR parameter set
    for key in _ignored_keys:
        if input_dict.get(key) is not None:
            wmess = f'automatically removing value of key {key}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict.pop(key)

    # get rid of structure related inputs that are overwritten from structure input
    for key in ['BRAVAIS', 'ALATBASIS', 'NAEZ', '<ZATOM>', '<RBASIS>', 'CARTESIAN']:
        if input_dict.get(key) is not None:
            wmess = f'automatically removing value of key {key}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict.pop(key)

    # automatically rescale RMAX, GMAX, RCLUSTZ, RCLUSTXY which are scaled with the lattice constant
    if alat_input is not None:
        if input_dict.get('RMAX') is not None:
            wmess = f'rescale RMAX: {alat_input / alat}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict['RMAX'] = input_dict['RMAX'] * alat_input / alat
        if input_dict.get('GMAX') is not None:
            wmess = f'rescale GMAX: {1 / (alat_input / alat)}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict['GMAX'] = input_dict['GMAX'] * 1 / (alat_input / alat)
        if input_dict.get('RCLUSTZ') is not None:
            wmess = f'rescale RCLUSTZ: {alat_input / alat}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict['RCLUSTZ'] = input_dict['RCLUSTZ'] * alat_input / alat
        if input_dict.get('RCLUSTXY') is not None:
            wmess = f'rescale RCLUSTXY: {alat_input / alat}'
            print('WARNING: ' + wmess)
            warnings.append(wmess)
            input_dict['RCLUSTXY'] = input_dict['RCLUSTXY'] * alat_input / alat

    # empty kkrparams instance (contains formatting info etc.)
    if not isvoronoi:
        params = kkrparams()
    else:
        params = kkrparams(params_type='voronoi')

    # for KKR calculation set EMIN automatically from parent_calc (always in res.emin of voronoi and kkr) if not provided in input node
    if (('EMIN' not in list(input_dict.keys()) or input_dict['EMIN'] is None) and parent_calc is not None):
        wmess = f'Overwriting EMIN with value from parent calculation {parent_calc}'
        print('WARNING: ' + wmess)
        warnings.append(wmess)
        if parent_calc.process_class == VoronoiCalculation:
            emin = parent_calc.outputs.output_parameters.get_dict().get('emin')
        else:
            emin = parent_calc.outputs.output_parameters.get_dict().get('energy_contour_group').get('emin')
        print('Setting emin:', emin, 'is emin None?', emin is None)
        params.set_value('EMIN', emin)

    # overwrite keywords with input parameter
    for key in list(input_dict.keys()):
        params.set_value(key, input_dict[key], silent=True)

    # Write input to file (the parameters that are set here are not allowed to be modfied externally)
    params.set_multiple_values(
        BRAVAIS=bravais, ALATBASIS=alat, NAEZ=naez, ZATOM=charges, RBASIS=positions, CARTESIAN=True
    )
    # for CPA case:
    if len(weights) > naez:
        natyp = len(weights)
        params.set_value('NATYP', natyp)
        params.set_value('<CPA-CONC>', weights)
        params.set_value('<SITE>', isitelist)
    else:
        natyp = naez

    # write shapes (extracted from voronoi parent automatically in kkr calculation plugin)
    if shapes is not None:
        params.set_value('<SHAPE>', shapes)

    # change input values of 2D input to new alat:
    rbl = params.get_value('<RBLEFT>')
    rbr = params.get_value('<RBRIGHT>')
    zper_l = params.get_value('ZPERIODL')
    zper_r = params.get_value('ZPERIODR')
    if rbl is not None:
        params.set_value('<RBLEFT>', array(rbl) * a_to_bohr / alat)
    if rbr is not None:
        params.set_value('<RBRIGHT>', array(rbr) * a_to_bohr / alat)
    if zper_l is not None:
        params.set_value('ZPERIODL', array(zper_l) * a_to_bohr / alat)
    if zper_r is not None:
        params.set_value('ZPERIODR', array(zper_r) * a_to_bohr / alat)

    # write inputfile
    params.fill_keywords_to_inputfile(output=input_filename)

    nspin = params.get_value('NSPIN')

    newsosol = False
    if 'NEWSOSOL' in params.get_value('RUNOPT'):
        newsosol = True

    return natyp, nspin, newsosol, warnings


def check_2Dinput_consistency(structure, parameters):
    """
    Check if structure and parameter data are complete and matching.

    :param input: structure, needs to be a valid aiida StructureData node
    :param input: parameters, needs to be valid aiida Dict node

    returns (False, errormessage) if an inconsistency has been found, otherwise return (True, '2D consistency check complete')
    """
    # default is bulk, get 2D info from structure.pbc info (periodic boundary contitions)
    is2D = False
    if not all(structure.pbc):
        # check periodicity, assumes finite size in z-direction
        if structure.pbc != (True, True, False):
            return (
                False,
                'Structure.pbc is neither (True, True, True) for bulk nor (True, True, False) for surface calculation!'
            )
        is2D = True

    # check for necessary info in 2D case
    inp_dict = parameters.get_dict()
    set_keys = [i for i in list(inp_dict.keys()) if inp_dict[i] is not None]
    has2Dinfo = True
    for icheck in ['INTERFACE', '<NRBASIS>', '<RBLEFT>', '<RBRIGHT>', 'ZPERIODL', 'ZPERIODR', '<NLBASIS>']:
        if icheck not in set_keys:
            has2Dinfo = False
    if has2Dinfo and not inp_dict['INTERFACE'] and is2D:
        return (False, "'INTERFACE' parameter set to False but structure is 2D")

    if has2Dinfo != is2D:
        if is2D:
            return (
                False,
                '2D info given in parameters but structure is 3D\nstructure is 2D? {}\ninput has 2D info? {}\nset keys are: {}'
                .format(is2D, has2Dinfo, set_keys)
            )
        return (
            False,
            '3D info given in parameters but structure is 2D\nstructure is 2D? {}\ninput has 2D info? {}\nset keys are: {}'
            .format(is2D, has2Dinfo, set_keys)
        )

    # if everything is ok:
    return (True, '2D consistency check complete')


def structure_from_params(parameters):
    """
    Construct aiida structure out of kkr parameter set (if ALATBASIS, RBASIS, ZATOM etc. are given)

    :param input: parameters, kkrparams object with structure information set (e.g. extracted from read_inputcard function)

    :returns: success, boolean to determine if structure creatoin was successful
    :returns: structure, an aiida StructureData object
    """
    from masci_tools.io.common_functions import get_aBohr2Ang
    from aiida.common.constants import elements as PeriodicTableElements
    from aiida.orm import StructureData
    from masci_tools.io.kkr_params import kkrparams
    from numpy import array

    # check input
    if not isinstance(parameters, kkrparams):
        raise InputValidationError('input parameters needs to be a "kkrparams" instance!')

    # initialize some stuff
    is_complete = True
    for icheck in ['<ZATOM>', '<RBASIS>', 'BRAVAIS', 'ALATBASIS']:
        if parameters.get_value(icheck) is None:
            is_complete = False

    # set natyp
    natyp = parameters.get_value('NATYP')
    naez = parameters.get_value('NAEZ')
    if natyp is None:
        if naez is None:
            is_complete = False
        else:
            natyp = naez

    # check if all necessary info for 2D calculation is there
    if parameters.get_value('INTERFACE'):
        for icheck in ['<NRBASIS>', '<RBLEFT>', '<RBRIGHT>', 'ZPERIODL', 'ZPERIODR', '<NLBASIS>']:
            if parameters.get_value(icheck) is None:
                is_complete = False

    # check CPA case
    if natyp != naez:
        for icheck in ['<SITE>', '<CPA-CONC>']:
            if parameters.get_value(icheck) is None:
                is_complete = False

    if not is_complete:
        return is_complete, StructureData()

    # extract cell using BRAVAIS and ALATBASIS and create empty structure
    alat = parameters.get_value('ALATBASIS')
    cell = array(parameters.get_value('BRAVAIS')) * alat * get_aBohr2Ang()
    struc = StructureData(cell=cell)

    # extract atom numbers
    zatom_all = parameters.get_value('<ZATOM>')

    # extract sites with positions, charges/Atom labels, weights
    # positions in units of alat
    pos_all = array(parameters.get_value('<RBASIS>'))
    if len(pos_all.shape) == 1:
        pos_all = array([pos_all])
        zatom_all = [zatom_all]
    if not parameters.get_value('CARTESIAN'):
        # convert from internal to cartesian coordinates
        for isite, tmp_pos in enumerate(pos_all):
            # cell already contains alat factor to convert to Ang. units
            pos_all[isite] = tmp_pos[0]*cell[0] + \
                tmp_pos[1]*cell[1] + tmp_pos[2]*cell[2]
    else:
        pos_all = pos_all * alat * get_aBohr2Ang()  # now positions are in Ang. units

    # convert to list if input contains a single entry only
    if not isinstance(zatom_all, list):
        zatom_all = [zatom_all]
        pos_all = [pos_all]

    # extract weights and sites for CPA calculations
    if natyp == naez:
        weights = [1. for i in range(natyp)]
        sites = list(range(1, natyp + 1))
    else:
        weights = parameters.get_value('<CPA-CONC>')
        sites = parameters.get_value('<SITE>')

    # fill structure from zatom, weights and sites information
    for isite in sites:
        pos = pos_all[sites.index(isite)]
        weight = weights[sites.index(isite)]
        if abs(zatom_all[isite - 1] - int(zatom_all[isite - 1])) > 10**-4:
            # TODO deal with VCA (non-integer zatom)
            print('VCA not implemented yet, stopping here!')
            raise NotImplementedError('VCA functionality not implemented')

        if zatom_all[isite - 1] < 1:
            symbol = 'X'
            struc.append_atom(position=pos, symbols='X', weights=weight)
        else:
            symbol = PeriodicTableElements.get(zatom_all[isite - 1]).get('symbol')
            struc.append_atom(position=pos, symbols=symbol, weights=weight)

    # set correct pbc for 2D case
    if parameters.get_value('INTERFACE'):
        struc.set_pbc((True, True, False))

    # finally return structure
    return is_complete, struc


def vca_check(structure, parameters):
    """

    """
    nsites = 0
    for site in structure.sites:
        sitekind = structure.get_kind(site.kind_name)
        nsites += len(sitekind.symbols)
    # VCA mode if CPAINFO = [-1,-1] first
    try:
        if parameters.get_dict().get('CPAINFO')[0] < 0:
            params_vca_mode = True
        else:
            params_vca_mode = False
    except:
        params_vca_mode = False
    # check if structure supports VCA mode
    vca_structure = False
    if params_vca_mode:
        if nsites > len(structure.sites):
            vca_structure = True

    return vca_structure


def get_username(computer):
    """
    set upload dir (get the remote username and try 5 times if there was a connection error
    """
    import time
    try_trans = 0
    while try_trans < 5:
        try_trans += 1
        try:
            with computer.get_transport() as transport:
                remote_user = transport.whoami()
        except:
            # this means we have some ssh connection error, thus we wait 5 seconds before we try again
            remote_user = None
            time.sleep(5)
        if remote_user is not None:
            break
    # check if username was extracted correctly and raise an error otherwise
    if remote_user is None:
        raise ValueError('Error getting the username from the computer!')

    return remote_user


def get_natyp(structure):
    """Count number of atom types (>NAEZ for CPA) for the structure"""
    counter = 0  # for CPA
    for site in structure.sites:
        sitekind = structure.get_kind(site.kind_name)
        for ikind in range(len(sitekind.symbols)):
            counter += 1
    return counter



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-KKR package.                                 #
#                                                                             #
# For further information on the license, see the LICENSE.txt file            #
###############################################################################
"""Command line utilities to create and inspect `StructureData` nodes."""
import click
from aiida.cmdline.params import options
from aiida.cmdline.utils import decorators, echo
from aiida.orm import Dict
from aiida_kkr.tools import kkrparams
from aiida_kkr.tools.common_workfunctions import structure_from_params
from . import cmd_data


@click.group('structure')
def cmd_structure():
    """Commands to create and inspect `StructureData` nodes."""


cmd_data.add_command(cmd_structure)


@cmd_structure.command('import')
@click.argument('filename', type=click.Path(exists=True))
@click.option(
    '--kkrpara/--no-kkrpara', default=False, show_default=True, help='Store also the kkrparams Dict in the db.'
)
@click.option('--verbose', '-v', is_flag=True, default=False, show_default=True, help='Print verbose output.')
@options.DRY_RUN()
@decorators.with_dbenv()
def cmd_import(filename, kkrpara, verbose, dry_run):
    """
    Import a `StructureData` from a KKRhost input file.

    FILENAME is the name/path of the inputcard file to use.

    If you want to import a structure from another file type you can use
    'verdi data structure import -ase <filename>' instead.
    """

    kkr_para = kkrparams()
    if verbose:
        print(f'start reading from file: {filename}')
    kkr_para.read_keywords_from_inputcard(filename)
    if verbose:
        print(f'read kkr params: {kkr_para.get_set_values()}')

    # extract kkr params
    missing = kkr_para.get_missing_keys(use_aiida=True)
    if kkrpara:
        if len(missing) > 0:
            # TODO replace by click echo
            echo.echo_warning(
                f"""inputcard did not contain all necessary keys to run a kkr calculation!
The following keys are missing:
{missing}"""
            )

        struc_keys = ['BRAVAIS', '<RBASIS>', '<ZATOM>', 'ALATBASIS', 'NAEZ', 'NATYP']
        no_struc_kkrparams = {k: v for k, v in kkr_para.get_set_values() if k not in struc_keys}
        if len(no_struc_kkrparams.keys()) == 0:
            echo.echo_critical('failed to extract kkr params')
        else:
            param_node = Dict(no_struc_kkrparams)
            if dry_run:
                echo.echo_success('parsed kkr params from input file')
            else:
                param_node.store()
                message = f'stored kkr params Dict node <{param_node.pk}>'
                echo.echo_success(message)

    # extract structure
    success, struc_node = structure_from_params(kkr_para)

    if success:
        formula = struc_node.get_formula()
        if dry_run:
            echo.echo_success(f'parsed structure from kkr params from input file with formula {formula}')
        else:
            struc_node.store()
            message = f'parsed and stored StructureData <{struc_node.pk}> with formula {formula}'
            echo.echo_success(message)
    else:
        echo.echo_critical('Error in getting the structure, check the input file.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is the calcfunction that parses the KKRhost DOS output, include interpolation of complex.dos files
"""

from aiida import orm
from aiida.engine import calcfunction
from masci_tools.io.common_functions import (interpolate_dos, get_Ry2eV)


@calcfunction
def parse_dosfiles(dos_retrieved):
    """
    parse dos files to XyData nodes
    take files from the retrieved folder of a KKRhost DOS run
    if BdG files are found we also parse and return those results
    """

    if 'complex.dos' in dos_retrieved.list_object_names():
        with dos_retrieved.open('complex.dos') as dosfolder:
            ef, dos, dos_int = interpolate_dos(dosfolder, return_original=True)
    else:
        raise ValueError('`complex.dos` not found in retrieved')

    dosnode, dosnode2 = make_XyData_nodes(dos, dos_int, ef)
    return_nodes = {'dos_data': dosnode, 'dos_data_interpol': dosnode2}

    for labeladd in ['_eh', '_he', '_hole']:
        # also try tp parse BdG output
        if 'complex.dos' + labeladd in dos_retrieved.list_object_names():
            with dos_retrieved.open('complex.dos' + labeladd) as dosfolder:
                ef, dos, dos_int = interpolate_dos(dosfolder, return_original=True)
            dosnode, dosnode2 = make_XyData_nodes(dos, dos_int, ef, labeladd)
            return_nodes['dos_data' + labeladd] = dosnode
            return_nodes['dos_data_interpol' + labeladd] = dosnode2

    return return_nodes


def make_XyData_nodes(dos, dos_int, ef, labeladd=''):
    """
    create XyData nodes from the dos and dos_int arrays, also transform to eV units before saving
    """
    eVscale = get_Ry2eV()
    # convert to eV units
    dos[:, :, 0] = (dos[:, :, 0] - ef) * eVscale
    dos[:, :, 1:] = dos[:, :, 1:] / eVscale
    dos_int[:, :, 0] = (dos_int[:, :, 0] - ef) * eVscale
    dos_int[:, :, 1:] = dos_int[:, :, 1:] / eVscale

    # create output nodes
    dosnode = orm.XyData()
    dosnode.set_x(dos[:, :, 0], 'E-EF', 'eV')
    name = ['tot', 's', 'p', 'd', 'f', 'g']
    name = name[:len(dos[0, 0, 1:]) - 1] + ['ns']
    ylists = [[], [], []]
    for line, _name in enumerate(name):
        ylists[0].append(dos[:, :, 1 + line])
        ylists[1].append(f'dos {_name}')
        ylists[2].append('states/eV')
    dosnode.set_y(ylists[0], ylists[1], ylists[2])
    dosnode.label = 'dos_data' + labeladd
    dosnode.description = 'Array data containing uniterpolated DOS (i.e. dos at finite imaginary part of energy). 3D array with (atoms, energy point, l-channel) dimensions.'

    # now create XyData node for interpolated data
    dosnode2 = orm.XyData()
    dosnode2.set_x(dos_int[:, :, 0], 'E-EF', 'eV')
    ylists = [[], [], []]
    for line, _name in enumerate(name):
        ylists[0].append(dos_int[:, :, 1 + line])
        ylists[1].append(f'interpolated dos {_name}')
        ylists[2].append('states/eV')
    dosnode2.set_y(ylists[0], ylists[1], ylists[2])
    dosnode2.label = 'dos_interpol_data' + labeladd
    dosnode2.description = 'Array data containing interpolated DOS (i.e. dos at real axis). 3D array with (atoms, energy point, l-channel) dimensions.'
    return dosnode, dosnode2
